"""Base class for encoders and generic multi encoders."""

import abc
import itertools
import six

import tensorflow as tf

from opennmt.layers.reducer import ConcatReducer, JoinReducer
from opennmt.utils import compat


@six.add_metaclass(abc.ABCMeta)
class Encoder(tf.keras.layers.Layer):
  """Base class for encoders."""

  def build_mask(self, inputs, sequence_length=None, dtype=tf.bool):
    """Builds a boolean mask for :obj:`inputs`."""
    if sequence_length is None:
      return None
    return tf.sequence_mask(sequence_length, maxlen=tf.shape(inputs)[1], dtype=dtype)

  def call(self, inputs, sequence_length=None, training=None):  # pylint: disable=arguments-differ
    """Encodes an input sequence.

    Args:
      inputs: The inputs to encode of shape :math:`[B, T, ...]`.
      sequence_length: The length of each input with shape :math:`[B]`.
      training: Run in training mode.

    Returns:
      A tuple ``(outputs, state, sequence_length)``.
    """
    return self.encode(
        inputs,
        sequence_length=sequence_length,
        mode=tf.estimator.ModeKeys.TRAIN if training else None)

  def encode(self, inputs, sequence_length=None, mode=tf.estimator.ModeKeys.TRAIN):
    """Encodes an input sequence.

    Args:
      inputs: The inputs to encode of shape :math:`[B, T, ...]`.
      sequence_length: The length of each input with shape :math:`[B]`.
      mode: A ``tf.estimator.ModeKeys`` mode.

    Returns:
      A tuple ``(outputs, state, sequence_length)``.
    """
    return self.call(
        inputs,
        sequence_length=sequence_length,
        training=mode == tf.estimator.ModeKeys.TRAIN)


class SequentialEncoder(Encoder):
  """An encoder that executes multiple encoders sequentially with optional
  transition layers.

  See for example "Cascaded Encoder" in https://arxiv.org/abs/1804.09849.
  """

  def __init__(self, encoders, states_reducer=JoinReducer(), transition_layer_fn=None):
    """Initializes the parameters of the encoder.

    Args:
      encoders: A list of :class:`opennmt.encoders.encoder.Encoder`.
      states_reducer: A :class:`opennmt.layers.reducer.Reducer` to merge all
        states.
      transition_layer_fn: A callable or list of callables applied to the
        output of an encoder before passing it as input to the next. If it is a
        single callable, it is applied between every encoders. Otherwise, the
        ``i`` th callable will be applied between encoders ``i`` and ``i + 1``.

    Raises:
      ValueError: if :obj:`transition_layer_fn` is a list with a size not equal
        to the number of encoder transitions ``len(encoders) - 1``.
    """
    if (transition_layer_fn is not None and isinstance(transition_layer_fn, list)
        and len(transition_layer_fn) != len(encoders) - 1):
      raise ValueError("The number of transition layers must match the number of encoder "
                       "transitions, expected %d layers but got %d."
                       % (len(encoders) - 1, len(transition_layer_fn)))
    super(SequentialEncoder, self).__init__()
    self.encoders = encoders
    self.states_reducer = states_reducer
    self.transition_layer_fn = transition_layer_fn

  def encode(self, inputs, sequence_length=None, mode=tf.estimator.ModeKeys.TRAIN):
    encoder_state = []

    for i, encoder in enumerate(self.encoders):
      with compat.tf_compat(v1="variable_scope")("encoder_{}".format(i)):
        if i > 0 and self.transition_layer_fn is not None:
          if isinstance(self.transition_layer_fn, list):
            inputs = self.transition_layer_fn[i - 1](inputs)
          else:
            inputs = self.transition_layer_fn(inputs)
        inputs, state, sequence_length = encoder.encode(
            inputs,
            sequence_length=sequence_length,
            mode=mode)
        encoder_state.append(state)

    return (
        inputs,
        self.states_reducer(encoder_state),
        sequence_length)


class ParallelEncoder(Encoder):
  """An encoder that encodes its input with several encoders and reduces the
  outputs and states together. Additional layers can be applied on each encoder
  output and on the combined output (e.g. to layer normalize each encoder
  output).

  If the input is a single ``tf.Tensor``, the same input will be encoded by
  every encoders. Otherwise, when the input is a Python sequence (e.g. the non
  reduced output of a :class:`opennmt.inputters.inputter.ParallelInputter`),
  each encoder will encode its corresponding input in the sequence.

  See for example "Multi-Columnn Encoder" in https://arxiv.org/abs/1804.09849.
  """

  def __init__(self,
               encoders,
               outputs_reducer=ConcatReducer(axis=1),
               states_reducer=JoinReducer(),
               outputs_layer_fn=None,
               combined_output_layer_fn=None,
               share_parameters=False):
    """Initializes the parameters of the encoder.

    Args:
      encoders: A list of :class:`opennmt.encoders.encoder.Encoder` or a single
        one, in which case the same encoder is applied to each input.
      outputs_reducer: A :class:`opennmt.layers.reducer.Reducer` to merge all
        outputs. If ``None``, defaults to
        :class:`opennmt.layers.reducer.JoinReducer`.
      states_reducer: A :class:`opennmt.layers.reducer.Reducer` to merge all
        states. If ``None``, defaults to
        :class:`opennmt.layers.reducer.JoinReducer`.
      outputs_layer_fn: A callable or list of callables applied to the
        encoders outputs If it is a single callable, it is on each encoder
        output. Otherwise, the ``i`` th callable is applied on encoder ``i``
        output.
      combined_output_layer_fn: A callable to apply on the combined output
        (i.e. the output of :obj:`outputs_reducer`).
      share_parameters: If ``True``, share parameters between the parallel
        encoders. For stateful encoders, simply pass a single encoder instance
        to :obj:`encoders` for parameter sharing.

    Raises:
      ValueError: if :obj:`outputs_layer_fn` is a list with a size not equal
        to the number of encoders.
    """
    if (isinstance(encoders, list)
        and outputs_layer_fn is not None and isinstance(outputs_layer_fn, list)
        and len(outputs_layer_fn) != len(encoders)):
      raise ValueError("The number of output layers must match the number of encoders; "
                       "expected %d layers but got %d."
                       % (len(encoders), len(outputs_layer_fn)))
    super(ParallelEncoder, self).__init__()
    self.encoders = encoders
    self.outputs_reducer = outputs_reducer if outputs_reducer is not None else JoinReducer()
    self.states_reducer = states_reducer if states_reducer is not None else JoinReducer()
    self.outputs_layer_fn = outputs_layer_fn
    self.combined_output_layer_fn = combined_output_layer_fn
    self.share_parameters = share_parameters

  def encode(self, inputs, sequence_length=None, mode=tf.estimator.ModeKeys.TRAIN):
    all_outputs = []
    all_states = []
    all_sequence_lengths = []
    parallel_inputs = isinstance(inputs, (list, tuple))
    parallel_encoders = isinstance(self.encoders, (list, tuple))

    if parallel_encoders and parallel_inputs and len(inputs) != len(self.encoders):
      raise ValueError("ParallelEncoder expects as many inputs as parallel encoders")
    if parallel_encoders:
      encoders = self.encoders
    else:
      encoders = itertools.repeat(self.encoders, len(inputs) if parallel_inputs else 1)

    for i, encoder in enumerate(encoders):
      scope_name = "encoder_{}".format(i) if not self.share_parameters else "parallel_encoder"
      reuse = self.share_parameters and i > 0
      with compat.tf_compat(v1="variable_scope")(scope_name, reuse=reuse):
        if parallel_inputs:
          encoder_inputs = inputs[i]
          length = sequence_length[i]
        else:
          encoder_inputs = inputs
          length = sequence_length

        outputs, state, length = encoder.encode(
            encoder_inputs,
            sequence_length=length,
            mode=mode)

        if self.outputs_layer_fn is not None:
          if isinstance(self.outputs_layer_fn, list):
            outputs = self.outputs_layer_fn[i](outputs)
          else:
            outputs = self.outputs_layer_fn(outputs)

        all_outputs.append(outputs)
        all_states.append(state)
        all_sequence_lengths.append(length)

    outputs, sequence_length = self.outputs_reducer(
        all_outputs, sequence_length=all_sequence_lengths)

    if self.combined_output_layer_fn is not None:
      outputs = self.combined_output_layer_fn(outputs)

    return (outputs, self.states_reducer(all_states), sequence_length)



# pylint: disable=W0223

"""Define RNN-based decoders."""

import inspect

import tensorflow as tf

from tensorflow.python.estimator.util import fn_args

from opennmt.decoders import decoder
from opennmt.utils.cell import build_cell
from opennmt.layers.reducer import align_in_time
from opennmt.layers.transformer import build_sequence_mask, multi_head_attention


class RNNDecoder(decoder.Decoder):
  """A basic RNN decoder."""

  def __init__(self,
               num_layers,
               num_units,
               bridge=None,
               cell_class=None,
               dropout=0.3,
               residual_connections=False):
    """Initializes the decoder parameters.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      bridge: A :class:`opennmt.layers.bridge.Bridge` to pass the encoder state
        to the decoder.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a LSTM cell.
      dropout: The probability to drop units in each layer output.
      residual_connections: If ``True``, each layer input will be added to its
        output.
    """
    self.num_layers = num_layers
    self.num_units = num_units
    self.bridge = bridge
    self.cell_class = cell_class
    self.dropout = dropout
    self.residual_connections = residual_connections

  @property
  def output_size(self):
    """Returns the decoder output size."""
    return self.num_units

  def _init_state(self, zero_state, initial_state=None):
    if initial_state is None:
      return zero_state
    elif self.bridge is None:
      raise ValueError("A bridge must be configured when passing encoder state")
    else:
      return self.bridge(initial_state, zero_state)

  def _get_attention(self, state, step=None):  # pylint: disable=unused-argument
    return None

  def _build_cell(self,
                  mode,
                  batch_size,
                  initial_state=None,
                  memory=None,
                  memory_sequence_length=None,
                  dtype=None):
    _ = memory_sequence_length

    if memory is None and dtype is None:
      raise ValueError("dtype argument is required when memory is not set")

    cell = build_cell(
        self.num_layers,
        self.num_units,
        mode,
        dropout=self.dropout,
        residual_connections=self.residual_connections,
        cell_class=self.cell_class)

    initial_state = self._init_state(
        cell.zero_state(batch_size, dtype or memory.dtype), initial_state=initial_state)

    return cell, initial_state

  def decode(self,
             inputs,
             sequence_length,
             vocab_size=None,
             initial_state=None,
             sampling_probability=None,
             embedding=None,
             output_layer=None,
             mode=tf.estimator.ModeKeys.TRAIN,
             memory=None,
             memory_sequence_length=None,
             return_alignment_history=False):
    _ = memory
    _ = memory_sequence_length

    batch_size = tf.shape(inputs)[0]

    if (sampling_probability is not None
        and (tf.contrib.framework.is_tensor(sampling_probability)
             or sampling_probability > 0.0)):
      if embedding is None:
        raise ValueError("embedding argument must be set when using scheduled sampling")

      tf.summary.scalar("sampling_probability", sampling_probability)
      helper = tf.contrib.seq2seq.ScheduledEmbeddingTrainingHelper(
          inputs,
          sequence_length,
          embedding,
          sampling_probability)
      fused_projection = False
    else:
      helper = tf.contrib.seq2seq.TrainingHelper(inputs, sequence_length)
      fused_projection = True  # With TrainingHelper, project all timesteps at once.

    cell, initial_state = self._build_cell(
        mode,
        batch_size,
        initial_state=initial_state,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        dtype=inputs.dtype)

    if output_layer is None:
      output_layer = decoder.build_output_layer(
          self.output_size, vocab_size, dtype=inputs.dtype)

    basic_decoder = tf.contrib.seq2seq.BasicDecoder(
        cell,
        helper,
        initial_state,
        output_layer=output_layer if not fused_projection else None)

    outputs, state, length = tf.contrib.seq2seq.dynamic_decode(basic_decoder)

    if fused_projection and output_layer is not None:
      logits = output_layer(outputs.rnn_output)
    else:
      logits = outputs.rnn_output
    # Make sure outputs have the same time_dim as inputs
    inputs_len = tf.shape(inputs)[1]
    logits = align_in_time(logits, inputs_len)

    if return_alignment_history:
      alignment_history = self._get_attention(state)
      if alignment_history is not None:
        alignment_history = align_in_time(alignment_history, inputs_len)
      return (logits, state, length, alignment_history)
    return (logits, state, length)

  def step_fn(self,
              mode,
              batch_size,
              initial_state=None,
              memory=None,
              memory_sequence_length=None,
              dtype=tf.float32):
    cell, initial_state = self._build_cell(
        mode,
        batch_size,
        initial_state=initial_state,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        dtype=dtype)

    def _fn(step, inputs, state, mode):
      _ = mode
      # This scope is defined by tf.contrib.seq2seq.dynamic_decode during the
      # training.
      with tf.variable_scope("decoder"):
        outputs, state = cell(inputs, state)
        if self.support_alignment_history:
          return outputs, state, self._get_attention(state, step=step)
        return outputs, state

    return _fn, initial_state


def _build_attention_mechanism(attention_mechanism,
                               num_units,
                               memory,
                               memory_sequence_length=None):
  """Builds an attention mechanism from a class or a callable."""
  if inspect.isclass(attention_mechanism):
    kwargs = {}
    if "dtype" in fn_args(attention_mechanism):
      # For TensorFlow 1.5+, dtype should be set in the constructor.
      kwargs["dtype"] = memory.dtype
    return attention_mechanism(
        num_units, memory, memory_sequence_length=memory_sequence_length, **kwargs)
  elif callable(attention_mechanism):
    return attention_mechanism(
        num_units, memory, memory_sequence_length)
  else:
    raise ValueError("Unable to build the attention mechanism")


class AttentionalRNNDecoder(RNNDecoder):
  """A RNN decoder with attention.

  It simple overrides the cell construction to add an attention wrapper.
  """

  def __init__(self,
               num_layers,
               num_units,
               bridge=None,
               attention_mechanism_class=None,
               output_is_attention=True,
               cell_class=None,
               dropout=0.3,
               residual_connections=False):
    """Initializes the decoder parameters.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      bridge: A :class:`opennmt.layers.bridge.Bridge` to pass the encoder state
        to the decoder.
      attention_mechanism_class: A class inheriting from
        ``tf.contrib.seq2seq.AttentionMechanism`` or a callable that takes
        ``(num_units, memory, memory_sequence_length)`` as arguments and returns
        a ``tf.contrib.seq2seq.AttentionMechanism``. Defaults to
        ``tf.contrib.seq2seq.LuongAttention``.
      output_is_attention: If ``True``, the final decoder output (before logits)
        is the output of the attention layer. In all cases, the output of the
        attention layer is passed to the next step.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell.
      dropout: The probability to drop units in each layer output.
      residual_connections: If ``True``, each layer input will be added to its
        output.
    """
    if attention_mechanism_class is None:
      attention_mechanism_class = tf.contrib.seq2seq.LuongAttention
    super(AttentionalRNNDecoder, self).__init__(
        num_layers,
        num_units,
        bridge=bridge,
        cell_class=cell_class,
        dropout=dropout,
        residual_connections=residual_connections)
    self.attention_mechanism_class = attention_mechanism_class
    self.output_is_attention = output_is_attention

  @property
  def support_alignment_history(self):
    return True

  def _get_attention(self, state, step=None):
    alignment_history = state.alignment_history
    if step is not None:
      return alignment_history.read(step)
    return tf.transpose(alignment_history.stack(), perm=[1, 0, 2])

  def _build_cell(self,
                  mode,
                  batch_size,
                  initial_state=None,
                  memory=None,
                  memory_sequence_length=None,
                  dtype=None):
    attention_mechanism = _build_attention_mechanism(
        self.attention_mechanism_class,
        self.num_units,
        memory,
        memory_sequence_length=memory_sequence_length)

    cell, initial_cell_state = RNNDecoder._build_cell(
        self,
        mode,
        batch_size,
        initial_state=initial_state,
        dtype=memory.dtype)

    cell = tf.contrib.seq2seq.AttentionWrapper(
        cell,
        attention_mechanism,
        attention_layer_size=self.num_units,
        alignment_history=True,
        output_attention=self.output_is_attention,
        initial_cell_state=initial_cell_state)

    if mode == tf.estimator.ModeKeys.TRAIN and self.dropout > 0.0:
      cell = tf.nn.rnn_cell.DropoutWrapper(
          cell, output_keep_prob=1.0 - self.dropout)

    initial_state = cell.zero_state(batch_size, memory.dtype)

    return cell, initial_state


class MultiAttentionalRNNDecoder(RNNDecoder):
  """A RNN decoder with multi-attention.

  This decoder can attend the encoder outputs after multiple RNN layers using
  one or multiple attention mechanisms. Additionally, the cell state of this
  decoder is not initialized from the encoder state (i.e. a
  :class:`opennmt.layers.bridge.ZeroBridge` is imposed).
  """

  def __init__(self,
               num_layers,
               num_units,
               attention_layers=None,
               attention_mechanism_class=None,
               cell_class=None,
               dropout=0.3,
               residual_connections=False):
    """Initializes the decoder parameters.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      attention_layers: A list of integers, the layers after which to add
        attention. If ``None``, attention will only be added after the last
        layer.
      attention_mechanism_class: A class or list of classes inheriting from
        ``tf.contrib.seq2seq.AttentionMechanism``. Alternatively, the class can
        be replaced by a callable that takes
        ``(num_units, memory, memory_sequence_length)`` as arguments and returns
        a ``tf.contrib.seq2seq.AttentionMechanism``. Defaults to
        ``tf.contrib.seq2seq.LuongAttention``.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell.
      dropout: The probability to drop units in each layer output.
      residual_connections: If ``True``, each layer input will be added to its
        output.
    """
    super(MultiAttentionalRNNDecoder, self).__init__(
        num_layers,
        num_units,
        cell_class=cell_class,
        dropout=dropout,
        residual_connections=residual_connections)

    attention_layers = attention_layers or [-1]
    attention_layers = [l % num_layers for l in attention_layers]
    if attention_mechanism_class is None:
      attention_mechanism_class = tf.contrib.seq2seq.LuongAttention
    if not isinstance(attention_mechanism_class, list):
      attention_mechanism_class = [attention_mechanism_class for _ in attention_layers]

    self.attention_mechanism_class = attention_mechanism_class
    self.attention_layers = attention_layers

  def _build_cell(self,
                  mode,
                  batch_size,
                  initial_state=None,
                  memory=None,
                  memory_sequence_length=None,
                  dtype=None):
    attention_mechanisms = [
        _build_attention_mechanism(
            attention_mechanism,
            self.num_units,
            memory,
            memory_sequence_length=memory_sequence_length)
        for attention_mechanism in self.attention_mechanism_class]

    cell = build_cell(
        self.num_layers,
        self.num_units,
        mode,
        dropout=self.dropout,
        residual_connections=self.residual_connections,
        cell_class=self.cell_class,
        attention_layers=self.attention_layers,
        attention_mechanisms=attention_mechanisms)

    initial_state = cell.zero_state(batch_size, memory.dtype)

    return cell, initial_state


class RNMTPlusDecoder(RNNDecoder):
  """The RNMT+ decoder described in https://arxiv.org/abs/1804.09849."""

  def __init__(self,
               num_layers,
               num_units,
               num_heads,
               cell_class=None,
               dropout=0.3):
    """Initializes the decoder parameters.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      num_heads: The number of attention heads.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a layer normalized LSTM cell.
      dropout: The probability to drop units from the decoder input and in each
        layer output.
    """
    if cell_class is None:
      cell_class = tf.contrib.rnn.LayerNormBasicLSTMCell
    super(RNMTPlusDecoder, self).__init__(
        num_layers,
        num_units,
        cell_class=cell_class,
        dropout=dropout)
    self.num_heads = num_heads

  @property
  def output_size(self):
    """Returns the decoder output size."""
    return self.num_units * 2

  def _build_cell(self,
                  mode,
                  batch_size,
                  initial_state=None,
                  memory=None,
                  memory_sequence_length=None,
                  dtype=None):
    cell = _RNMTPlusDecoderCell(
        mode,
        self.num_layers,
        self.num_units,
        self.num_heads,
        memory,
        memory_sequence_length,
        cell_class=self.cell_class,
        dropout=self.dropout)
    return cell, cell.zero_state(batch_size, dtype)

class _RNMTPlusDecoderCell(tf.nn.rnn_cell.RNNCell):

  def __init__(self,
               mode,
               num_layers,
               num_units,
               num_heads,
               memory,
               memory_sequence_length,
               cell_class=None,
               dropout=0.3):
    super(_RNMTPlusDecoderCell, self).__init__()
    self._mode = mode
    self._num_units = num_units
    self._num_heads = num_heads
    self._dropout = dropout
    self._cells = [cell_class(num_units) for _ in range(num_layers)]
    self._memory = memory
    self._memory_mask = build_sequence_mask(
        memory_sequence_length,
        num_heads=self._num_heads,
        maximum_length=tf.shape(memory)[1])

  @property
  def state_size(self):
    return tuple(cell.state_size for cell in self._cells)

  @property
  def output_size(self):
    return self._num_units * 2

  def zero_state(self, batch_size, dtype):
    with tf.name_scope("RNMTPlusDecoderCellZeroState", values=[batch_size]):
      return tuple(cell.zero_state(batch_size, dtype) for cell in self._cells)

  def __call__(self, inputs, state, scope=None):
    inputs = tf.layers.dropout(
        inputs, rate=self._dropout, training=self._mode == tf.estimator.ModeKeys.TRAIN)

    new_states = []
    with tf.variable_scope("rnn_0"):
      last_outputs, state_0 = self._cells[0](inputs, state[0])
      new_states.append(state_0)

    with tf.variable_scope("multi_head_attention"):
      context = multi_head_attention(
          self._num_heads,
          tf.expand_dims(last_outputs, 1),
          self._memory,
          self._mode,
          mask=self._memory_mask,
          dropout=self._dropout)
      context = tf.squeeze(context, axis=1)

    for i in range(1, len(self._cells)):
      inputs = tf.concat([last_outputs, context], axis=-1)
      with tf.variable_scope("rnn_%d" % i):
        outputs, state_i = self._cells[i](inputs, state[i])
        new_states.append(state_i)
        outputs = tf.layers.dropout(
            outputs, rate=self._dropout, training=self._mode == tf.estimator.ModeKeys.TRAIN)
        if i >= 2:
          outputs += last_outputs
        last_outputs = outputs

    final = tf.concat([last_outputs, context], -1)
    return final, tuple(new_states)



# coding=utf-8
# Copyright 2018 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Changes:
# * raise exception on incompatible TensorFlow version
# * fix Pylint warnings

"""Optimizer variants which make it possible to use very large batch sizes with
limited GPU memory. Optimizers in this module accumulate the gradients for n
batches, and call the optimizer's update rule every n batches with the
accumulated gradients.
See [Saunders et al., 2018](https://arxiv.org/abs/1805.00456) for details.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports

import tensorflow as tf


class MultistepAdamOptimizer(tf.train.AdamOptimizer):
  """Adam with SGD updates every n steps with accumulated gradients."""

  def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, epsilon=1e-8,
               use_locking=False, name="Adam", n=1):
    super(MultistepAdamOptimizer, self).__init__(
        learning_rate=learning_rate, beta1=beta1, beta2=beta2, epsilon=epsilon,
        use_locking=use_locking, name=name)
    if not hasattr(self, "_create_non_slot_variable"):
      raise RuntimeError("MultistepAdamOptimizer requires TensorFlow 1.6+")
    self._n = n  # Call Adam optimizer every n batches with accumulated grads
    self._n_t = None  # n as tensor

  def _create_slots(self, var_list):
    """Create slot variables for Adam with accumulated gradients.

    Like super class method, but additionally creates slots for the gradient
    accumulator `acc_grad` and the counter variable.
    """
    super(MultistepAdamOptimizer, self)._create_slots(var_list)
    first_var = min(var_list, key=lambda x: x.name)
    self._create_non_slot_variable(initial_value=0 if self._n == 1 else 1,
                                   name="iter",
                                   colocate_with=first_var)
    for v in var_list:
      self._zeros_slot(v, "grad_acc", self._name)

  def _get_iter_variable(self):
    if tf.contrib.eager.in_eager_mode():
      graph = None
    else:
      graph = tf.get_default_graph()
    return self._get_non_slot_variable("iter", graph=graph)

  def _prepare(self):
    super(MultistepAdamOptimizer, self)._prepare()
    self._n_t = tf.convert_to_tensor(self._n, name="n")

  def _apply_cond(self, apply_fn, grad, var, *args, **kwargs):
    """Conditionally apply or accumulate gradient.

    Call `apply_fn only if the current counter value (iter) is zero. This
    method couples common functionality for all _apply_*() implementations
    in Adam.
    """
    grad_acc = self.get_slot(var, "grad_acc")

    def _apply_adam(grad_acc, apply_fn, grad, var, *args, **kwargs):
      total_grad = (grad_acc + grad) / tf.cast(self._n_t, grad.dtype)
      adam_op = apply_fn(total_grad, var, *args, **kwargs)
      with tf.control_dependencies([adam_op]):
        grad_acc_to_zero_op = grad_acc.assign(tf.zeros_like(grad_acc),
                                              use_locking=self._use_locking)
      return tf.group(adam_op, grad_acc_to_zero_op)

    def _accumulate_gradient(grad_acc, grad):
      assign_op = tf.assign_add(grad_acc, grad, use_locking=self._use_locking)
      return tf.group(assign_op)  # Strip return value

    return tf.cond(tf.equal(self._get_iter_variable(), 0),
                   lambda: _apply_adam(grad_acc, apply_fn, grad, var, *args, **kwargs),
                   lambda: _accumulate_gradient(grad_acc, grad))

  def _apply_dense(self, grad, var):
    return self._apply_cond(
        super(MultistepAdamOptimizer, self)._apply_dense, grad, var)

  def _resource_apply_dense(self, grad, var):
    return self._apply_cond(
        super(MultistepAdamOptimizer, self)._resource_apply_dense, grad, var)

  def _apply_sparse_shared(self, grad, var, indices, scatter_add):
    return self._apply_cond(
        super(MultistepAdamOptimizer, self)._apply_sparse_shared, grad, var,
        indices, scatter_add)

  def _apply_sparse(self, grad, var):
    # TODO: Implement a sparse version
    dense_grad = tf.convert_to_tensor(grad)
    return self._apply_cond(
        super(MultistepAdamOptimizer, self)._apply_dense, dense_grad, var)

  def _finish(self, update_ops, name_scope):
    """Like super class method, but updates beta_power variables only every
    n batches. The iter variable is updated with

       iter <- iter + 1 mod n
    """
    iter_ = self._get_iter_variable()
    beta1_power, beta2_power = self._get_beta_accumulators()
    with tf.control_dependencies(update_ops):
      with tf.colocate_with(iter_):

        def _update_beta_op():
          update_beta1 = beta1_power.assign(
              beta1_power * self._beta1_t,
              use_locking=self._use_locking)
          update_beta2 = beta2_power.assign(
              beta2_power * self._beta2_t,
              use_locking=self._use_locking)
          return tf.group(update_beta1, update_beta2)
        maybe_update_beta = tf.cond(tf.equal(iter_, 0), _update_beta_op, tf.no_op)
        with tf.control_dependencies([maybe_update_beta]):
          update_iter = iter_.assign(tf.mod(iter_ + 1, self._n_t),
                                     use_locking=self._use_locking)
    return tf.group(
        *update_ops + [update_iter, maybe_update_beta], name=name_scope)



"""Base class and functions for dynamic decoders."""

import abc
import six

import tensorflow as tf

from opennmt.utils import beam_search


def logits_to_cum_log_probs(logits, sequence_length):
  """Returns the cumulated log probabilities of sequences.

  Args:
    logits: The sequence of logits of shape :math:`[B, T, ...]`.
    sequence_length: The length of each sequence of shape :math:`[B]`.

  Returns:
    The cumulated log probability of each sequence.
  """
  mask = tf.sequence_mask(
      sequence_length, maxlen=tf.shape(logits)[1], dtype=logits.dtype)
  mask = tf.expand_dims(mask, -1)

  log_probs = tf.nn.log_softmax(logits)
  log_probs = log_probs * mask
  log_probs = tf.reduce_max(log_probs, axis=-1)
  log_probs = tf.reduce_sum(log_probs, axis=1)

  return log_probs

def get_embedding_fn(embedding):
  """Returns the embedding function.

  Args:
    embedding: The embedding tensor or a callable that takes word ids.

  Returns:
    A callable that takes word ids.
  """
  if callable(embedding):
    return embedding
  else:
    return lambda ids: tf.nn.embedding_lookup(embedding, ids)

def build_output_layer(num_units, vocab_size, dtype=None):
  """Builds the output projection layer.

  Args:
    num_units: The layer input depth.
    vocab_size: The layer output depth.
    dtype: The layer dtype.

  Returns:
    A ``tf.layers.Dense`` instance.

  Raises:
    ValueError: if :obj:`vocab_size` is ``None``.
  """
  if vocab_size is None:
    raise ValueError("vocab_size must be set to build the output layer")

  layer = tf.layers.Dense(vocab_size, use_bias=True, dtype=dtype)
  layer.build([None, num_units])
  return layer

def get_sampling_probability(global_step,
                             read_probability=None,
                             schedule_type=None,
                             k=None):
  """Returns the sampling probability as described in
  https://arxiv.org/abs/1506.03099.

  Args:
    global_step: The training step.
    read_probability: The probability to read from the inputs.
    schedule_type: The type of schedule.
    k: The convergence constant.

  Returns:
    The probability to sample from the output ids as a 0D ``tf.Tensor`` or
    ``None`` if scheduled sampling is not configured.

  Raises:
    ValueError: if :obj:`schedule_type` is set but not :obj:`k` or if
     :obj:`schedule_type` is ``linear`` but an initial :obj:`read_probability`
     is not set.
    TypeError: if :obj:`schedule_type` is invalid.
  """
  if read_probability is None and schedule_type is None:
    return None

  if schedule_type is not None and schedule_type != "constant":
    if k is None:
      raise ValueError("scheduled_sampling_k is required when scheduled_sampling_type is set")

    step = tf.cast(global_step, tf.float32)
    k = tf.constant(k, tf.float32)

    if schedule_type == "linear":
      if read_probability is None:
        raise ValueError("Linear schedule requires an initial read probability")
      read_probability = min(read_probability, 1.0)
      read_probability = tf.maximum(read_probability - k * step, 0.0)
    elif schedule_type == "exponential":
      read_probability = tf.pow(k, step)
    elif schedule_type == "inverse_sigmoid":
      read_probability = k / (k + tf.exp(step / k))
    else:
      raise TypeError("Unknown scheduled sampling type: {}".format(schedule_type))

  return 1.0 - read_probability


@six.add_metaclass(abc.ABCMeta)
class Decoder(object):
  """Base class for decoders."""

  def decode(self,
             inputs,
             sequence_length,
             vocab_size=None,
             initial_state=None,
             sampling_probability=None,
             embedding=None,
             output_layer=None,
             mode=tf.estimator.ModeKeys.TRAIN,
             memory=None,
             memory_sequence_length=None,
             return_alignment_history=False):
    """Decodes a full input sequence.

    Usually used for training and evaluation where target sequences are known.

    Args:
      inputs: The input to decode of shape :math:`[B, T, ...]`.
      sequence_length: The length of each input with shape :math:`[B]`.
      vocab_size: The output vocabulary size. Must be set if :obj:`output_layer`
        is not set.
      initial_state: The initial state as a (possibly nested tuple of...) tensors.
      sampling_probability: The probability of sampling categorically from
        the output ids instead of reading directly from the inputs.
      embedding: The embedding tensor or a callable that takes word ids.
        Must be set when :obj:`sampling_probability` is set.
      output_layer: Optional layer to apply to the output prior sampling.
        Must be set if :obj:`vocab_size` is not set.
      mode: A ``tf.estimator.ModeKeys`` mode.
      memory: (optional) Memory values to query.
      memory_sequence_length: (optional) Memory values length.
      return_alignment_history: If ``True``, also returns the alignment
        history from the attention layer (``None`` will be returned if
        unsupported by the decoder).

    Returns:
      A tuple ``(outputs, state, sequence_length)`` or
      ``(outputs, state, sequence_length, alignment_history)``
      if :obj:`return_alignment_history` is ``True``.
    """
    _ = embedding
    if sampling_probability is not None:
      raise ValueError("Scheduled sampling is not supported by this decoder")
    if (not self.support_multi_source
        and memory is not None
        and tf.contrib.framework.nest.is_sequence(memory)):
      raise ValueError("Multiple source encodings are passed to this decoder "
                       "but it does not support multi source context. You should "
                       "instead configure your encoder to merge the different "
                       "encodings.")

    returned_values = self.decode_from_inputs(
        inputs,
        sequence_length,
        initial_state=initial_state,
        mode=mode,
        memory=memory,
        memory_sequence_length=memory_sequence_length)

    if self.support_alignment_history:
      outputs, state, attention = returned_values
    else:
      outputs, state = returned_values
      attention = None

    if output_layer is None:
      output_layer = build_output_layer(self.output_size, vocab_size, dtype=inputs.dtype)
    logits = output_layer(outputs)
    if return_alignment_history:
      return (logits, state, tf.identity(sequence_length), attention)
    return (logits, state, tf.identity(sequence_length))

  @property
  def support_alignment_history(self):
    """Returns ``True`` if this decoder can return the attention as alignment
    history."""
    return False

  @property
  def support_multi_source(self):
    """Returns ``True`` if this decoder supports multiple source context."""
    return False

  def dynamic_decode(self,
                     embedding,
                     start_tokens,
                     end_token,
                     vocab_size=None,
                     initial_state=None,
                     output_layer=None,
                     maximum_iterations=250,
                     minimum_length=0,
                     mode=tf.estimator.ModeKeys.PREDICT,
                     memory=None,
                     memory_sequence_length=None,
                     dtype=None,
                     return_alignment_history=False,
                     sample_from=1,
                     sample_temperature=1):
    """Decodes dynamically from :obj:`start_tokens` with greedy search.

    Usually used for inference.

    Args:
      embedding: The embedding tensor or a callable that takes word ids.
      start_tokens: The start token ids with shape :math:`[B]`.
      end_token: The end token id.
      vocab_size: The output vocabulary size. Must be set if :obj:`output_layer`
        is not set.
      initial_state: The initial state as a (possibly nested tuple of...) tensors.
      output_layer: Optional layer to apply to the output prior sampling.
        Must be set if :obj:`vocab_size` is not set.
      maximum_iterations: The maximum number of decoding iterations.
      minimum_length: The minimum length of decoded sequences (:obj:`end_token`
        excluded).
      mode: A ``tf.estimator.ModeKeys`` mode.
      memory: (optional) Memory values to query.
      memory_sequence_length: (optional) Memory values length.
      dtype: The data type. Required if :obj:`memory` is ``None``.
      return_alignment_history: If ``True``, also returns the alignment
        history from the attention layer (``None`` will be returned if
        unsupported by the decoder).
      sample_from: Sample predictions from the :obj:`sample_from` most likely
        tokens. If 0, sample from the full output distribution.
      sample_temperature: Value dividing logits. In random sampling, a high
        value generates more random samples.

    Returns:
      A tuple ``(predicted_ids, state, sequence_length, log_probs)`` or
      ``(predicted_ids, state, sequence_length, log_probs, alignment_history)``
      if :obj:`return_alignment_history` is ``True``.
    """
    return Decoder.dynamic_decode_and_search(
        self,
        embedding,
        start_tokens,
        end_token,
        vocab_size=vocab_size,
        initial_state=initial_state,
        output_layer=output_layer,
        beam_width=1,
        length_penalty=0.0,
        maximum_iterations=maximum_iterations,
        minimum_length=minimum_length,
        mode=mode,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        dtype=dtype,
        return_alignment_history=return_alignment_history,
        sample_from=sample_from,
        sample_temperature=sample_temperature)

  def dynamic_decode_and_search(self,
                                embedding,
                                start_tokens,
                                end_token,
                                vocab_size=None,
                                initial_state=None,
                                output_layer=None,
                                beam_width=5,
                                length_penalty=0.0,
                                maximum_iterations=250,
                                minimum_length=0,
                                mode=tf.estimator.ModeKeys.PREDICT,
                                memory=None,
                                memory_sequence_length=None,
                                dtype=None,
                                return_alignment_history=False,
                                sample_from=1,
                                sample_temperature=1):
    """Decodes dynamically from :obj:`start_tokens` with beam search.

    Usually used for inference.

    Args:
      embedding: The embedding tensor or a callable that takes word ids.
      start_tokens: The start token ids with shape :math:`[B]`.
      end_token: The end token id.
      vocab_size: The output vocabulary size. Must be set if :obj:`output_layer`
        is not set.
      initial_state: The initial state as a (possibly nested tuple of...) tensors.
      output_layer: Optional layer to apply to the output prior sampling.
        Must be set if :obj:`vocab_size` is not set.
      beam_width: The width of the beam.
      length_penalty: The length penalty weight during beam search.
      maximum_iterations: The maximum number of decoding iterations.
      minimum_length: The minimum length of decoded sequences (:obj:`end_token`
        excluded).
      mode: A ``tf.estimator.ModeKeys`` mode.
      memory: (optional) Memory values to query.
      memory_sequence_length: (optional) Memory values length.
      dtype: The data type. Required if :obj:`memory` is ``None``.
      return_alignment_history: If ``True``, also returns the alignment
        history from the attention layer (``None`` will be returned if
        unsupported by the decoder).
      sample_from: Sample predictions from the :obj:`sample_from` most likely
        tokens. If 0, sample from the full output distribution.
      sample_temperature: Value dividing logits. In random sampling, a high
        value generates more random samples.

    Returns:
      A tuple ``(predicted_ids, state, sequence_length, log_probs)`` or
      ``(predicted_ids, state, sequence_length, log_probs, alignment_history)``
      if :obj:`return_alignment_history` is ``True``.
    """
    if sample_from != 1 and beam_width > 1:
      raise ValueError("Sampling decoding is not compatible with beam search, "
                       "set beam_width to 1 instead.")
    batch_size = tf.shape(start_tokens)[0] * beam_width
    if dtype is None:
      if memory is None:
        raise ValueError("dtype argument is required when no memory is set")
      dtype = tf.contrib.framework.nest.flatten(memory)[0].dtype

    if beam_width > 1:
      if initial_state is not None:
        initial_state = tf.contrib.seq2seq.tile_batch(initial_state, multiplier=beam_width)
      if memory is not None:
        memory = tf.contrib.seq2seq.tile_batch(memory, multiplier=beam_width)
      if memory_sequence_length is not None:
        memory_sequence_length = tf.contrib.seq2seq.tile_batch(
            memory_sequence_length, multiplier=beam_width)

    embedding_fn = get_embedding_fn(embedding)
    step_fn, initial_state = self.step_fn(
        mode,
        batch_size,
        initial_state=initial_state,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        dtype=dtype)
    if output_layer is None:
      if vocab_size is None:
        raise ValueError("vocab_size must be known when the output_layer is not set")
      output_layer = build_output_layer(self.output_size, vocab_size, dtype=dtype)

    state = {"decoder": initial_state}
    if self.support_alignment_history and not tf.contrib.framework.nest.is_sequence(memory):
      state["attention"] = tf.zeros([batch_size, 0, tf.shape(memory)[1]], dtype=dtype)

    def _symbols_to_logits_fn(ids, step, state):
      if ids.shape.ndims == 2:
        ids = ids[:, -1]
      inputs = embedding_fn(ids)
      returned_values = step_fn(step, inputs, state["decoder"], mode)
      if self.support_alignment_history:
        outputs, state["decoder"], attention = returned_values
        if "attention" in state:
          state["attention"] = tf.concat([state["attention"], tf.expand_dims(attention, 1)], 1)
      else:
        outputs, state["decoder"] = returned_values
      logits = output_layer(outputs)
      return logits, state

    if beam_width == 1:
      outputs, lengths, log_probs, state = greedy_decode(
          _symbols_to_logits_fn,
          start_tokens,
          end_token,
          decode_length=maximum_iterations,
          state=state,
          return_state=True,
          min_decode_length=minimum_length,
          last_step_as_input=True,
          sample_from=sample_from,
          sample_temperature=sample_temperature)
    else:
      outputs, log_probs, state = beam_search.beam_search(
          _symbols_to_logits_fn,
          start_tokens,
          beam_width,
          maximum_iterations,
          vocab_size,
          length_penalty,
          states=state,
          eos_id=end_token,
          return_states=True,
          tile_states=False,
          min_decode_length=minimum_length)
      lengths = tf.not_equal(outputs, 0)
      lengths = tf.cast(lengths, tf.int32)
      lengths = tf.reduce_sum(lengths, axis=-1) - 1  # Ignore </s>
      outputs = outputs[:, :, 1:]  # Ignore <s>.

    attention = state.get("attention")
    if beam_width == 1:
      # Make shape consistent with beam search.
      outputs = tf.expand_dims(outputs, 1)
      lengths = tf.expand_dims(lengths, 1)
      log_probs = tf.expand_dims(log_probs, 1)
      if attention is not None:
        attention = tf.expand_dims(attention, 1)
    if attention is not None:
      attention = attention[:, :, 1:]  # Ignore attention for <s>.

    if return_alignment_history:
      return (outputs, state["decoder"], lengths, log_probs, attention)
    return (outputs, state["decoder"], lengths, log_probs)

  def decode_from_inputs(self,
                         inputs,
                         sequence_length,
                         initial_state=None,
                         mode=tf.estimator.ModeKeys.TRAIN,
                         memory=None,
                         memory_sequence_length=None):
    """Decodes from full inputs.

    Args:
      inputs: The input to decode of shape :math:`[B, T, ...]`.
      sequence_length: The length of each input with shape :math:`[B]`.
      initial_state: The initial state as a (possibly nested tuple of...) tensors.
      mode: A ``tf.estimator.ModeKeys`` mode.
      memory: (optional) Memory values to query.
      memory_sequence_length: (optional) Memory values length.

    Returns:
      A tuple ``(outputs, state)`` or ``(outputs, state, attention)``
      if ``self.support_alignment_history``.
    """
    raise NotImplementedError()

  def step_fn(self,
              mode,
              batch_size,
              initial_state=None,
              memory=None,
              memory_sequence_length=None,
              dtype=tf.float32):
    """Callable to run decoding steps.

    Args:
      mode: A ``tf.estimator.ModeKeys`` mode.
      batch_size: The batch size.
      initial_state: The initial state to start from as a (possibly nested tuple
        of...) tensors.
      memory: (optional) Memory values to query.
      memory_sequence_length: (optional) Memory values length.
      dtype: The data type.

    Returns:
      A callable with the signature
      ``(step, inputs, state, mode)`` -> ``(outputs, state)`` or
      ``(outputs, state, attention)`` if ``self.support_alignment_history``.
    """
    raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class DecoderV2(tf.keras.layers.Layer):
  """Base class for decoders.

  Note:
    TensorFlow 2.0 version.
  """

  def __init__(self, num_sources=1, **kwargs):
    """Initializes the decoder parameters.

    Args:
      num_sources: The number of source contexts expected by this decoder.
      **kwargs: Additional layer arguments.

    Raises:
      ValueError: if the number of source contexts :obj:`num_sources` is not
        supported by this decoder.
    """
    if num_sources < self.minimum_sources or num_sources > self.maximum_sources:
      raise ValueError("This decoder accepts between %d and %d source contexts, "
                       "but received %d" % (
                           self.minimum_sources, self.maximum_sources, num_sources))
    super(DecoderV2, self).__init__(**kwargs)
    self.num_sources = num_sources
    self.output_layer = None

  @property
  def minimum_sources(self):
    """The minimum number of source contexts supported by this decoder."""
    return 1

  @property
  def maximum_sources(self):
    """The maximum number of source contexts supported by this decoder."""
    return 1

  def initialize(self, vocab_size=None, output_layer=None):
    """Initializes the decoder configuration.

    Args:
      vocab_size: The target vocabulary size.
      output_layer: The output layer to use.

    Raises:
      ValueError: if both :obj:`vocab_size` and :obj:`output_layer` are not set.
    """
    if output_layer is not None:
      self.output_layer = output_layer
    else:
      if vocab_size is None:
        raise ValueError("One of vocab_size and output_layer must be set")
      self.output_layer = tf.keras.layers.Dense(vocab_size, name="logits")

  def get_initial_state(self, initial_state=None, batch_size=None, dtype=tf.float32):
    """Returns the initial decoder state.

    Args:
      initial_state: An initial state to start from, e.g. the last encoder
        state.
      batch_size: The batch size to use.
      dtype: The dtype of the state.

    Returns:
      A nested structure of tensors representing the decoder state.

    Raises:
      RuntimeError: if the decoder was not initialized.
      ValueError: if one of :obj:`batch_size` or :obj:`dtype` is not set but
        :obj:`initial_state` is not passed.
    """
    self._assert_is_initialized()
    if batch_size is None or dtype is None:
      if initial_state is None:
        raise ValueError("Either initial_state should be set, or batch_size and "
                         "dtype should be set")
      first_state = tf.nest.flatten(initial_state)[0]
      if batch_size is None:
        batch_size = tf.shape(first_state)[0]
      if dtype is None:
        dtype = first_state.dtype
    return self._get_initial_state(batch_size, dtype, initial_state=initial_state)

  # pylint: disable=arguments-differ
  def call(self,
           inputs,
           length_or_step=None,
           state=None,
           memory=None,
           memory_sequence_length=None,
           training=None):
    """Runs the decoder layer on either a complete sequence (e.g. for training
    or scoring), or a single timestep (e.g. for iterative decoding).

    Args:
      inputs: The inputs to decode, can be a 3D (training) or 2D (iterative
        decoding) tensor.
      length_or_step: For 3D :obj:`inputs`, the length of each sequence. For 2D
        :obj:`inputs`, the current decoding timestep.
      state: The decoder state.
      memory: Memory values to query.
      memory_sequence_length: Memory values length.
      training: Run in training mode.

    Returns:
      A tuple with the logits, the decoder state, and an attention vector.

    Raises:
      RuntimeError: if the decoder was not initialized.
      ValueError: if the :obj:`inputs` rank is different than 2 or 3.
      ValueError: if :obj:`length_or_step` is invalid.
      ValueError: if the number of source contexts (:obj:`memory`) does not
        match the number defined at the decoder initialization.
    """
    self._assert_is_initialized()
    self._assert_memory_is_compatible(memory, memory_sequence_length)
    rank = inputs.shape.ndims
    if rank == 2:
      if length_or_step.shape.ndims != 0:
        raise ValueError("length_or_step should be a scalar with the current timestep")
      outputs, state, attention = self.step(
          inputs,
          length_or_step,
          state=state,
          memory=memory,
          memory_sequence_length=memory_sequence_length,
          training=training)
    elif rank == 3:
      if length_or_step.shape.ndims != 1:
        raise ValueError("length_or_step should contain the length of each sequence")
      outputs, state, attention = self.forward(
          inputs,
          sequence_length=length_or_step,
          initial_state=state,
          memory=memory,
          memory_sequence_length=memory_sequence_length,
          training=training)
    else:
      raise ValueError("Unsupported input rank %d" % rank)
    logits = self.output_layer(outputs)
    return logits, state, attention

  def forward(self,
              inputs,
              sequence_length=None,
              initial_state=None,
              memory=None,
              memory_sequence_length=None,
              training=None):
    """Runs the decoder on full sequences.

    Args:
      inputs: The 3D decoder input.
      sequence_length: The length of each input sequence.
      initial_state: The initial decoder state.
      memory: Memory values to query.
      memory_sequence_length: Memory values length.
      training: Run in training mode.

    Returns:
      A tuple with the decoder outputs, the decoder state, and the attention
      vector.
    """
    # TODO: implement this using step().
    raise NotImplementedError()

  @abc.abstractmethod
  def step(self,
           inputs,
           timestep,
           state=None,
           memory=None,
           memory_sequence_length=None,
           training=None):
    """Runs one decoding step.

    Args:
      inputs: The 2D decoder input.
      timestep: The current decoding step.
      state: The decoder state.
      memory: Memory values to query.
      memory_sequence_length: Memory values length.
      training: Run in training mode.

    Returns:
      A tuple with the decoder outputs, the decoder state, and the attention
      vector.
    """
    raise NotImplementedError()

  def _assert_is_initialized(self):
    """Raises an expection if the decoder was not initialized."""
    if self.output_layer is None:
      raise RuntimeError("The decoder was not initialized")

  def _assert_memory_is_compatible(self, memory, memory_sequence_length):
    """Raises an expection if the memory layout is not compatible with this decoder."""

    def _num_elements(obj):
      if obj is None:
        return 0
      elif isinstance(obj, (list, tuple)):
        return len(obj)
      else:
        return 1

    num_memory = _num_elements(memory)
    num_length = _num_elements(memory_sequence_length)
    if num_memory != num_length and memory_sequence_length is not None:
      raise ValueError("got %d memory values but %d length vectors" % (num_memory, num_length))
    if num_memory != self.num_sources:
      raise ValueError("expected %d source contexts, but got %d" % (
          self.num_sources, num_memory))


def greedy_decode(symbols_to_logits_fn,
                  initial_ids,
                  end_id,
                  decode_length=None,
                  state=None,
                  return_state=False,
                  min_decode_length=0,
                  last_step_as_input=False,
                  sample_from=1,
                  sample_temperature=1):
  """Greedily decodes from :obj:`initial_ids`.

  Args:
    symbols_to_logits_fn: Interface to the model, to provide logits.
        Shoud take [batch_size, decoded_ids] and return [batch_size, vocab_size].
    initial_ids: Ids to start off the decoding, this will be the first thing
        handed to symbols_to_logits_fn.
    eos_id: ID for end of sentence.
    decode_length: Maximum number of steps to decode for (EOS included).
    states: A dictionnary of (possibly nested) decoding states.
    return_state: If ``True``, also return the updated decoding states.
    min_decode_length: Minimum length of decoded hypotheses (EOS excluded).
    last_step_as_input: If ``True``, only feed the last predicted ids into
      :obj:`symbols_to_logits_fn`.
    sample_from: Sample from the :obj:`sample_from` most likely tokens. If 0,
      sample from the full output distribution.
    sample_temperature: Value dividing logits. In random sampling, a high value
      generates more random samples.

  Returns:
    A tuple with the decoded output, the decoded lengths, the log probabilities,
    and the decoding states (if :obj:`return_state` is ``True``).
  """
  batch_size = tf.shape(initial_ids)[0]
  batch_ids = tf.range(tf.cast(batch_size, initial_ids.dtype))

  def _condition(step, finished, unused_inputs, unused_outputs,
                 unused_lengths, unused_cum_log_probs, unused_state):
    cond = tf.logical_not(tf.reduce_all(finished))
    if decode_length is not None:
      cond = tf.logical_and(cond, step < decode_length)
    return cond

  def _body(step, finished, inputs, outputs, lengths, cum_log_probs, state):
    # Run next step.
    logits, state = symbols_to_logits_fn(inputs, step, state)
    logits = tf.cast(logits, tf.float32)
    if sample_temperature != 1:
      logits /= tf.cast(sample_temperature, logits.dtype)
    log_probs = tf.nn.log_softmax(logits)
    if min_decode_length > 0:
      log_probs = tf.cond(
          step < min_decode_length,
          true_fn=lambda: beam_search.penalize_token(log_probs, end_id),
          false_fn=lambda: log_probs)

    if sample_from == 1:  # Sample best prediction.
      sampled_ids = tf.argmax(log_probs, axis=-1, output_type=inputs.dtype)
    elif sample_from == 0:  # Sample from the full output distribution.
      distribution = tf.distributions.Categorical(probs=tf.exp(log_probs), dtype=inputs.dtype)
      sampled_ids = distribution.sample()
    else:  # Sample from the top K.
      topk_log_probs, topk_ids = tf.nn.top_k(log_probs, k=sample_from)
      topk_ids = tf.cast(topk_ids, inputs.dtype)
      distribution = tf.distributions.Categorical(logits=topk_log_probs, dtype=inputs.dtype)
      topk_sampled_ids = distribution.sample()
      sampled_ids = tf.gather_nd(topk_ids, tf.stack([batch_ids, topk_sampled_ids], axis=-1))

    sampled_log_probs = tf.gather_nd(log_probs, tf.stack([batch_ids, sampled_ids], axis=-1))
    outputs = outputs.write(step, sampled_ids)

    # Don't update finished batches.
    lengths += 1 - tf.cast(finished, lengths.dtype)
    cum_log_probs += sampled_log_probs * (1.0 - tf.cast(finished, sampled_log_probs.dtype))
    finished = tf.logical_or(finished, tf.equal(sampled_ids, end_id))
    if last_step_as_input:
      next_inputs = sampled_ids
    else:
      next_inputs = tf.concat([inputs, tf.expand_dims(sampled_ids, 1)], axis=1)
    return step + 1, finished, next_inputs, outputs, lengths, cum_log_probs, state

  step = tf.constant(0)
  finished = tf.zeros([batch_size], dtype=tf.bool)
  outputs = tf.TensorArray(initial_ids.dtype, size=0, dynamic_size=True)
  lengths = tf.zeros([batch_size], dtype=tf.int32)
  cum_log_probs = tf.zeros([batch_size], dtype=tf.float32)
  if not last_step_as_input:
    initial_ids = tf.expand_dims(initial_ids, 1)

  _, _, _, outputs, lengths, cum_log_probs, state = tf.while_loop(
      _condition,
      _body,
      loop_vars=(step, finished, initial_ids, outputs, lengths, cum_log_probs, state),
      shape_invariants=(
          step.get_shape(),
          finished.get_shape(),
          tf.TensorShape([None] if last_step_as_input else [None, None]),
          tf.TensorShape(None),
          lengths.get_shape(),
          cum_log_probs.get_shape(),
          tf.contrib.framework.nest.map_structure(
              beam_search.get_state_shape_invariants, state)),
      parallel_iterations=1)

  outputs = tf.transpose(outputs.stack())
  if return_state:
    return outputs, lengths, cum_log_probs, state
  return outputs, lengths, cum_log_probs



"""Module defining inputters.

Inputters implement the logic of transforming raw data to vectorized inputs,
e.g., from a line of text to a sequence of embeddings.
"""

from opennmt.inputters.inputter import Inputter, ParallelInputter, ExampleInputter, MixedInputter
from opennmt.inputters.text_inputter import TextInputter
from opennmt.inputters.text_inputter import WordEmbedder
from opennmt.inputters.text_inputter import CharConvEmbedder
from opennmt.inputters.text_inputter import CharRNNEmbedder
from opennmt.inputters.record_inputter import SequenceRecordInputter, write_sequence_record



"""Module defining custom optimizers."""

from opennmt.utils.compat import is_tf2

if not is_tf2():
  from opennmt.optimizers.adafactor import AdafactorOptimizer
  from opennmt.optimizers.adafactor import get_optimizer_from_params \
      as get_adafactor_optimizer_from_params

  from opennmt.optimizers.multistep_adam import MultistepAdamOptimizer
  from opennmt.optimizers.mixed_precision_wrapper import MixedPrecisionOptimizerWrapper



# -*- coding: utf-8 -*-

"""Noise layers."""

import abc
import six

import tensorflow as tf

from opennmt import constants


class WordNoiser(object):
  """Applies noise to words sequences."""

  def __init__(self, noises=None, subword_token="￭", is_spacer=False):
    """Initializes the noising class.

    Args:
      noises: A list of :class:`opennmt.layers.noise.Noise` instances to apply
        sequentially.
      subword_token: The special token used by the subword tokenizer. This is
        required when the noise should be applied at the word level and not the
        subword level.
      is_spacer: Whether :obj:`subword_token` is used as a spacer (as in
        SentencePiece) or a joiner (as in BPE).
    """
    if noises is None:
      noises = []
    self.noises = noises
    self.subword_token = subword_token
    self.is_spacer = is_spacer

  def add(self, noise):
    """Adds a noise to apply."""
    self.noises.append(noise)

  def __call__(self, tokens, sequence_length=None, keep_shape=True):
    """Applies noise on :obj:`tokens`.

    Args:
      tokens: A string ``tf.Tensor`` or batch of string ``tf.Tensor``.
      sequence_length: When :obj:`tokens` is 2D, the length of each sequence in
        the batch.
      keep_shape: Ensure that the shape is kept. Otherwise, fit the shape to the
        new lengths.

    Returns:
      A tuple with the noisy version of :obj:`tokens` and the new lengths.
    """
    rank = tokens.shape.ndims
    if rank == 1:
      input_length = tf.shape(tokens)[0]
      if sequence_length is not None:
        tokens = tokens[:sequence_length]
      else:
        tokens = tokens[:tf.math.count_nonzero(tokens)]
      words = tokens_to_words(
          tokens,
          subword_token=self.subword_token,
          is_spacer=self.is_spacer)
      for noise in self.noises:
        words = noise(words)
      outputs = tf.RaggedTensor.from_tensor(words, padding="").flat_values
      output_length = tf.shape(outputs)[0]
      if keep_shape:
        outputs = tf.pad(outputs, [[0, input_length - output_length]])
      return outputs, output_length
    elif rank == 2:
      if sequence_length is None:
        raise ValueError("sequence_length must be passed for 2D inputs")
      tokens, sequence_length = tf.map_fn(
          lambda arg: self(*arg, keep_shape=True),
          (tokens, sequence_length),
          back_prop=False)
      if not keep_shape:
        tokens = tokens[:, :tf.reduce_max(sequence_length)]
      return tokens, sequence_length
    else:
      raise NotImplementedError("unsupported input rank %d" % rank)


@six.add_metaclass(abc.ABCMeta)
class Noise(object):
  """Base class for noise modules."""

  def __call__(self, words):
    """Applies noise on a sequence of words.

    Args:
      words: The sequence of words as a string ``tf.Tensor``. If it has 2
        dimensions, each row represents a word that possibly contains multiple
        tokens.

    Returns:
      A noisy version of :obj:`words`.

    Raises:
      ValueError: if :obj:`words` has a rank greater than 2.
    """
    if words.shape.ndims > 2:
      raise ValueError("Noise only supports tensors of rank 2 or less")
    inputs = words
    if words.shape.ndims == 1:
      inputs = tf.expand_dims(inputs, 1)
    outputs = self._apply(inputs)
    if words.shape.ndims == 1:
      outputs = tf.squeeze(outputs, 1)
    return outputs

  @abc.abstractmethod
  def _apply(self, words):
    """Applies noise on a sequence of words.

    Args:
      words: A 2D string ``tf.Tensor`` where each row represents a word that
        possibly contains multiple tokens.

    Returns:
      A noisy version of :obj:`words`.
    """
    raise NotImplementedError()


class WordDropout(Noise):
  """Randomly drops words in a sequence."""

  def __init__(self, dropout):
    """Initializes the noise module.

    Args:
      dropout: The probability to drop word.
    """
    self.dropout = dropout

  def _apply(self, words):
    if self.dropout == 0:
      return tf.identity(words)
    num_words = tf.shape(words, out_type=tf.int64)[0]
    keep_mask = random_mask([num_words], 1 - self.dropout)
    keep_ind = tf.where(keep_mask)
    # Keep at least one word.
    keep_ind = tf.cond(
        tf.equal(tf.shape(keep_ind)[0], 0),
        true_fn=lambda: tf.random.uniform([1], maxval=num_words - 1, dtype=tf.int64),
        false_fn=lambda: tf.squeeze(keep_ind, -1))
    return tf.gather(words, keep_ind)


class WordReplacement(Noise):
  """Randomly replaces words."""

  def __init__(self, probability, filler=constants.UNKNOWN_TOKEN):
    """Initializes the noise module.

    Args:
      probability: The probability to replace words.
      filler: The replacement token.
    """
    self.probability = probability
    self.filler = filler

  def _apply(self, words):
    if self.probability == 0:
      return tf.identity(words)
    shape = tf.shape(words)
    replace_mask = random_mask(shape[:1], self.probability)
    filler = tf.fill([shape[0], 1], self.filler)
    filler = tf.pad(filler, [[0, 0], [0, shape[-1] - 1]])
    return tf.where(replace_mask, x=filler, y=words)


class WordPermutation(Noise):
  """Randomly permutes words in a sequence with a maximum distance."""

  def __init__(self, max_distance):
    """Initializes the noise module.

    Args:
      max_distance: The maximum permutation distance.
    """
    self.max_distance = max_distance

  def _apply(self, words):
    if self.max_distance == 0:
      return tf.identity(words)
    num_words = tf.shape(words)[0]
    offset = tf.random.uniform([num_words], maxval=1) * (self.max_distance + 1)
    offset = tf.cast(offset, num_words.dtype)
    new_pos = tf.argsort(tf.range(num_words) + offset)
    return tf.gather(words, new_pos)


def tokens_to_words(tokens, subword_token="￭", is_spacer=False):
  """Converts a sequence of tokens to a sequence of words.

  For example, if a BPE tokenization produces this sequence:

      ["He@@", "llo", "W@@", "orld", "@@!"]

  this function will return the tensor:

      [["He@@", "llo", ""], ["W@@", "orld", "@@!"]]

  Args:
    tokens: A 1D string ``tf.Tensor``.
    subword_token: The special token used by the subword tokenizer.
    is_spacer: Whether :obj:`subword_token` is used as a spacer (as in
      SentencePiece) or a joiner (as in BPE).

  Returns:
    A 2D string ``tf.Tensor``.
  """
  if is_spacer:
    subword = tf.strings.regex_full_match(tokens, "[^%s].*" % subword_token)
  else:
    right = tf.strings.regex_full_match(tokens, ".*%s" % subword_token)
    left = tf.strings.regex_full_match(tokens, "%s.*" % subword_token)
    subword = tf.logical_or(tf.roll(right, shift=1, axis=0), left)
  start = tf.logical_not(subword)
  start_indices = tf.squeeze(tf.where(start), -1)
  words = tf.RaggedTensor.from_row_starts(tokens, start_indices)
  return words.to_tensor()

def random_mask(shape, probability):
  """Generates a random boolean mask.

  Args:
    shape: The mask shape.
    probability: The probability to select an element.

  Returns:
    A boolean mask with shape :obj:`shape`.
  """
  probs = tf.random.uniform(shape, maxval=1)
  mask = tf.math.less(probs, probability)
  return mask



"""Define word-based embedders."""

import abc
import collections
import os
import six
import yaml

import numpy as np
import tensorflow as tf

from tensorflow.python.ops import lookup_ops as lookup
try:
  from tensorflow.contrib.tensorboard.plugins import projector
except ModuleNotFoundError:
  from tensorboard.plugins import projector

from google.protobuf import text_format

from opennmt import constants, tokenizers
from opennmt.inputters.inputter import Inputter
from opennmt.utils import compat
from opennmt.utils.cell import build_cell, last_encoding_from_state
from opennmt.utils.misc import count_lines
from opennmt.constants import PADDING_TOKEN


def visualize_embeddings(log_dir, embedding_var, vocabulary_file, num_oov_buckets=1):
  """Registers an embedding variable for visualization in TensorBoard.

  This function registers :obj:`embedding_var` in the ``projector_config.pbtxt``
  file and generates metadata from :obj:`vocabulary_file` to attach a label
  to each word ID.

  Args:
    log_dir: The active log directory.
    embedding_var: The embedding variable to visualize.
    vocabulary_file: The associated vocabulary file.
    num_oov_buckets: The number of additional unknown tokens.
  """
  # Copy vocabulary file to log_dir.
  basename = "%s.txt" % embedding_var.op.name.replace("/", "_")
  destination = os.path.join(log_dir, basename)
  tf.gfile.Copy(vocabulary_file, destination, overwrite=True)

  # Append <unk> tokens.
  with tf.gfile.Open(destination, mode="ab") as vocab:
    if num_oov_buckets == 1:
      vocab.write(b"<unk>\n")
    else:
      for i in range(num_oov_buckets):
        vocab.write(tf.compat.as_bytes("<unk%d>\n" % i))

  config = projector.ProjectorConfig()

  # If the projector file exists, load it.
  target = os.path.join(log_dir, "projector_config.pbtxt")
  if tf.gfile.Exists(target):
    with tf.gfile.Open(target, mode="rb") as target_file:
      text_format.Merge(target_file.read(), config)

  # If this embedding is already registered, just update the metadata path.
  exists = False
  for meta in config.embeddings:
    if meta.tensor_name == embedding_var.name:
      meta.metadata_path = basename
      exists = True
      break

  if not exists:
    embedding = config.embeddings.add()
    embedding.tensor_name = embedding_var.name
    embedding.metadata_path = basename

  summary_writer = tf.summary.FileWriter(log_dir)

  projector.visualize_embeddings(summary_writer, config)

def load_pretrained_embeddings(embedding_file,
                               vocabulary_file,
                               num_oov_buckets=1,
                               with_header=True,
                               case_insensitive_embeddings=True):
  """Returns pretrained embeddings relative to the vocabulary.

  The :obj:`embedding_file` must have the following format:

  .. code-block:: text

      N M
      word1 val1 val2 ... valM
      word2 val1 val2 ... valM
      ...
      wordN val1 val2 ... valM

  or if :obj:`with_header` is ``False``:

  .. code-block:: text

      word1 val1 val2 ... valM
      word2 val1 val2 ... valM
      ...
      wordN val1 val2 ... valM

  This function will iterate on each embedding in :obj:`embedding_file` and
  assign the pretrained vector to the associated word in :obj:`vocabulary_file`
  if found. Otherwise, the embedding is ignored.

  If :obj:`case_insensitive_embeddings` is ``True``, word embeddings are assumed
  to be trained on lowercase data. In that case, word alignments are case
  insensitive meaning the pretrained word embedding for "the" will be assigned
  to "the", "The", "THE", or any other case variants included in
  :obj:`vocabulary_file`.

  Args:
    embedding_file: Path the embedding file. Entries will be matched against
      :obj:`vocabulary_file`.
    vocabulary_file: The vocabulary file containing one word per line.
    num_oov_buckets: The number of additional unknown tokens.
    with_header: ``True`` if the embedding file starts with a header line like
      in GloVe embedding files.
    case_insensitive_embeddings: ``True`` if embeddings are trained on lowercase
      data.

  Returns:
    A Numpy array of shape ``[vocabulary_size + num_oov_buckets, embedding_size]``.
  """
  # Map words to ids from the vocabulary.
  word_to_id = collections.defaultdict(list)
  with compat.gfile_open(vocabulary_file, mode="rb") as vocabulary:
    count = 0
    for word in vocabulary:
      word = word.strip()
      if case_insensitive_embeddings:
        word = word.lower()
      word_to_id[word].append(count)
      count += 1

  # Fill pretrained embedding matrix.
  with compat.gfile_open(embedding_file, mode="rb") as embedding:
    pretrained = None

    if with_header:
      next(embedding)

    for line in embedding:
      fields = line.strip().split()
      word = fields[0]

      if pretrained is None:
        pretrained = np.random.normal(
            size=(count + num_oov_buckets, len(fields) - 1))

      # Lookup word in the vocabulary.
      if word in word_to_id:
        ids = word_to_id[word]
        for index in ids:
          pretrained[index] = np.asarray(fields[1:])

  return pretrained

def tokens_to_chars(tokens, padding_value=PADDING_TOKEN):
  """Splits tokens into unicode characters.

  Args:
    tokens: A string ``tf.Tensor`` of shape :math:`[T]`.
    padding_value: The value to use for padding.

  Returns:
    The characters as a string ``tf.Tensor`` of shape :math:`[T, W]` and the
    length of each token as an int64 ``tf.Tensor``  of shape :math:`[T]`.
  """
  if hasattr(tf, "strings") and hasattr(tf.strings, "unicode_split"):
    ragged = tf.strings.unicode_split(tokens, "UTF-8")
    chars = ragged.to_tensor(default_value=padding_value)
    lengths = ragged.row_lengths()
    return chars, lengths

  def _split_chars(token, max_length, delimiter=" "):
    chars = list(token.decode("utf-8"))
    while len(chars) < max_length:
      chars.append(padding_value)
    return delimiter.join(chars).encode("utf-8")

  def _string_len(token):
    return len(token.decode("utf-8"))

  def _apply():
    # Get the length of each token.
    lengths = tf.map_fn(
        lambda x: tf.py_func(_string_len, [x], tf.int64),
        tokens,
        dtype=tf.int64,
        back_prop=False)

    max_length = tf.reduce_max(lengths)

    # Add a delimiter between each unicode character.
    spaced_chars = tf.map_fn(
        lambda x: tf.py_func(_split_chars, [x, max_length], [tf.string]),
        tokens,
        dtype=[tf.string],
        back_prop=False)

    # Split on this delimiter
    chars = tf.map_fn(
        lambda x: tf.string_split(x, delimiter=" ").values,
        spaced_chars,
        dtype=tf.string,
        back_prop=False)

    return chars, lengths

  def _none():
    chars = tf.constant([], dtype=tf.string)
    lengths = tf.constant([], dtype=tf.int64)
    return chars, lengths

  chars, lengths = tf.cond(tf.equal(tf.shape(tokens)[0], 0), true_fn=_none, false_fn=_apply)
  chars.set_shape([None, None])
  lengths.set_shape([None])
  return chars, lengths

def _get_field(config, key, prefix=None, default=None, required=False):
  if prefix:
    key = "%s%s" % (prefix, key)
  value = config.get(key, default)
  if value is None and required:
    raise ValueError("Missing field '%s' in the data configuration" % key)
  return value


@six.add_metaclass(abc.ABCMeta)
class TextInputter(Inputter):
  """An abstract inputter that processes text."""

  def __init__(self,
               tokenizer=None,
               dtype=tf.float32,
               vocabulary_file_key=None,
               num_oov_buckets=1):
    super(TextInputter, self).__init__(dtype=dtype)
    self.tokenizer = tokenizer
    self.vocabulary_file_key = vocabulary_file_key
    self.num_oov_buckets = num_oov_buckets
    self.vocabulary = None
    self.vocabulary_size = None
    self.vocabulary_file = None

  def initialize(self, metadata, asset_dir=None, asset_prefix=""):
    self.vocabulary_file = metadata[self.vocabulary_file_key]
    self.vocabulary_size = count_lines(self.vocabulary_file) + self.num_oov_buckets
    if self.tokenizer is None:
      tokenizer_config = _get_field(metadata, "tokenization", prefix=asset_prefix)
      if tokenizer_config:
        if isinstance(tokenizer_config, six.string_types) and compat.gfile_exists(tokenizer_config):
          with compat.gfile_open(tokenizer_config, mode="rb") as config_file:
            tokenizer_config = yaml.load(config_file)
        self.tokenizer = tokenizers.OpenNMTTokenizer(params=tokenizer_config)
      else:
        self.tokenizer = tokenizers.SpaceTokenizer()
    self.tokenizer.initialize(metadata)
    return super(TextInputter, self).initialize(
        metadata, asset_dir=asset_dir, asset_prefix=asset_prefix)

  def export_assets(self, asset_dir, asset_prefix=""):
    return self.tokenizer.export_assets(asset_dir, asset_prefix=asset_prefix)

  def vocabulary_lookup(self):
    """Returns a lookup table mapping string to index."""
    return lookup.index_table_from_file(
        self.vocabulary_file,
        vocab_size=self.vocabulary_size - self.num_oov_buckets,
        num_oov_buckets=self.num_oov_buckets)

  def vocabulary_lookup_reverse(self):
    """Returns a lookup table mapping index to string."""
    return lookup.index_to_string_table_from_file(
        self.vocabulary_file,
        vocab_size=self.vocabulary_size - self.num_oov_buckets,
        default_value=constants.UNKNOWN_TOKEN)

  def make_dataset(self, data_file, training=None):
    self.vocabulary = self.vocabulary_lookup()
    return tf.data.TextLineDataset(data_file)

  def get_dataset_size(self, data_file):
    return count_lines(data_file)

  def make_features(self, element=None, features=None, training=None):
    """Tokenizes raw text."""
    if features is None:
      features = {}
    if "tokens" in features:
      return features
    tokens = self.tokenizer.tokenize(element)
    features["length"] = tf.shape(tokens)[0]
    features["tokens"] = tokens
    return features


class WordEmbedder(TextInputter):
  """Simple word embedder."""

  def __init__(self,
               vocabulary_file_key,
               embedding_size=None,
               embedding_file_key=None,
               embedding_file_with_header=True,
               case_insensitive_embeddings=True,
               trainable=True,
               dropout=0.0,
               tokenizer=None,
               dtype=tf.float32):
    """Initializes the parameters of the word embedder.

    Args:
      vocabulary_file_key: The data configuration key of the vocabulary file
        containing one word per line.
      embedding_size: The size of the resulting embedding.
        If ``None``, an embedding file must be provided.
      embedding_file_key: The data configuration key of the embedding file.
      embedding_file_with_header: ``True`` if the embedding file starts with a
        header line like in GloVe embedding files.
      case_insensitive_embeddings: ``True`` if embeddings are trained on
        lowercase data.
      trainable: If ``False``, do not optimize embeddings.
      dropout: The probability to drop units in the embedding.
      tokenizer: An optional :class:`opennmt.tokenizers.tokenizer.Tokenizer` to
        tokenize the input text. Defaults to a space tokenization.
      dtype: The embedding type.

    Raises:
      ValueError: if neither :obj:`embedding_size` nor :obj:`embedding_file_key`
        are set.

    See Also:
      The :meth:`opennmt.inputters.text_inputter.load_pretrained_embeddings`
      function for details about the pretrained embedding format and behavior.
    """
    super(WordEmbedder, self).__init__(
        tokenizer=tokenizer,
        dtype=dtype,
        vocabulary_file_key=vocabulary_file_key)
    self.embedding = None
    self.embedding_size = embedding_size
    self.embedding_file = None
    self.embedding_file_key = embedding_file_key
    self.embedding_file_with_header = embedding_file_with_header
    self.case_insensitive_embeddings = case_insensitive_embeddings
    self.trainable = trainable
    self.dropout = dropout

  def initialize(self, metadata, asset_dir=None, asset_prefix=""):
    assets = super(WordEmbedder, self).initialize(
        metadata, asset_dir=asset_dir, asset_prefix=asset_prefix)
    if self.embedding_file_key is not None:
      self.embedding_file = metadata[self.embedding_file_key]
    else:
      embedding = _get_field(metadata, "embedding", prefix=asset_prefix)
      if embedding is None and self.embedding_size is None:
        raise ValueError("embedding_size must be set")
      if embedding is not None:
        self.embedding_file = embedding["path"]
        self.trainable = embedding.get("trainable", True)
        self.embedding_file_with_header = embedding.get("with_header", True)
        self.case_insensitive_embeddings = embedding.get("case_insensitive", True)
    if self.embedding_file is None and self.embedding_size is None:
      raise ValueError("Must either provide embedding_size or embedding_file_key")

    return assets

  def get_receiver_tensors(self):
    return {
        "tokens": tf.placeholder(tf.string, shape=(None, None)),
        "length": tf.placeholder(tf.int32, shape=(None,))
    }

  def make_features(self, element=None, features=None, training=None):
    """Converts words tokens to ids."""
    features = super(WordEmbedder, self).make_features(
        element=element, features=features, training=training)
    if "ids" in features:
      return features
    if self.vocabulary is None:
      self.vocabulary = self.vocabulary_lookup()
    ids = self.vocabulary.lookup(features["tokens"])
    if not self.is_target:
      features["ids"] = ids
    else:
      bos = tf.constant([constants.START_OF_SENTENCE_ID], dtype=ids.dtype)
      eos = tf.constant([constants.END_OF_SENTENCE_ID], dtype=ids.dtype)
      features["ids"] = tf.concat([bos, ids], axis=0)
      features["ids_out"] = tf.concat([ids, eos], axis=0)
      features["length"] += 1 # Increment length accordingly.
    return features

  def build(self, input_shape=None):
    if self.embedding_file:
      pretrained = load_pretrained_embeddings(
          self.embedding_file,
          self.vocabulary_file,
          num_oov_buckets=self.num_oov_buckets,
          with_header=self.embedding_file_with_header,
          case_insensitive_embeddings=self.case_insensitive_embeddings)
      self.embedding_size = pretrained.shape[-1]
      initializer = tf.constant_initializer(value=pretrained.astype(self.dtype))
    else:
      initializer = tf.keras.initializers.glorot_uniform()
    shape = [self.vocabulary_size, self.embedding_size]
    self.embedding = tf.Variable(
        initial_value=lambda: initializer(shape, dtype=self.dtype),
        trainable=self.trainable,
        name=compat.name_from_variable_scope("w_embs"))
    super(WordEmbedder, self).build(input_shape)

  def make_inputs(self, features, training=None):
    if not self.built:
      self.build()
    outputs = tf.nn.embedding_lookup(self.embedding, features["ids"])
    if training and self.dropout > 0:
      outputs = tf.keras.layers.Dropout(self.dropout)(outputs, training=training)
    return outputs

  def visualize(self, log_dir):
    visualize_embeddings(
        log_dir,
        self.embedding,
        self.vocabulary_file,
        num_oov_buckets=self.num_oov_buckets)


@six.add_metaclass(abc.ABCMeta)
class CharEmbedder(TextInputter):
  """Base class for character-aware inputters."""

  def __init__(self,
               vocabulary_file_key,
               embedding_size,
               dropout=0.0,
               tokenizer=None,
               dtype=tf.float32):
    """Initializes the parameters of the character embedder.

    Args:
      vocabulary_file_key: The meta configuration key of the vocabulary file
        containing one character per line.
      embedding_size: The size of the character embedding.
      dropout: The probability to drop units in the embedding.
      tokenizer: An optional :class:`opennmt.tokenizers.tokenizer.Tokenizer` to
        tokenize the input text. Defaults to a space tokenization.
      dtype: The embedding type.
    """
    super(CharEmbedder, self).__init__(
        tokenizer=tokenizer,
        dtype=dtype,
        vocabulary_file_key=vocabulary_file_key)
    self.embedding_size = embedding_size
    self.embedding = None
    self.dropout = dropout

  def get_receiver_tensors(self):
    return {
        "chars": tf.placeholder(tf.string, shape=(None, None, None)),
        "length": tf.placeholder(tf.int32, shape=(None,))
    }

  def make_features(self, element=None, features=None, training=None):
    """Converts words to characters."""
    if features is None:
      features = {}
    if "char_ids" in features:
      return features
    if "chars" in features:
      chars = features["chars"]
    else:
      features = super(CharEmbedder, self).make_features(
          element=element, features=features, training=training)
      chars, _ = tokens_to_chars(features["tokens"])
    if self.vocabulary is None:
      self.vocabulary = self.vocabulary_lookup()
    features["char_ids"] = self.vocabulary.lookup(chars)
    return features

  def build(self, input_shape=None):
    shape = [self.vocabulary_size, self.embedding_size]
    initializer = tf.keras.initializers.glorot_uniform()
    self.embedding = tf.Variable(
        initial_value=lambda: initializer(shape, dtype=self.dtype),
        name=compat.name_from_variable_scope("w_char_embs"))
    super(CharEmbedder, self).build(input_shape)

  @abc.abstractmethod
  def make_inputs(self, features, training=None):
    raise NotImplementedError()

  def visualize(self, log_dir):
    visualize_embeddings(
        log_dir,
        self.embedding,
        self.vocabulary_file,
        num_oov_buckets=self.num_oov_buckets)

  def _embed(self, inputs, training):
    if not self.built:
      self.build()
    outputs = tf.nn.embedding_lookup(self.embedding, inputs)
    outputs = tf.layers.dropout(outputs, rate=self.dropout, training=training)
    return outputs


class CharConvEmbedder(CharEmbedder):
  """An inputter that applies a convolution on characters embeddings."""

  def __init__(self,
               vocabulary_file_key,
               embedding_size,
               num_outputs,
               kernel_size=5,
               stride=3,
               dropout=0.0,
               tokenizer=None,
               dtype=tf.float32):
    """Initializes the parameters of the character convolution embedder.

    Args:
      vocabulary_file_key: The meta configuration key of the vocabulary file
        containing one character per line.
      embedding_size: The size of the character embedding.
      num_outputs: The dimension of the convolution output space.
      kernel_size: Length of the convolution window.
      stride: Length of the convolution stride.
      dropout: The probability to drop units in the embedding.
      tokenizer: An optional :class:`opennmt.tokenizers.tokenizer.Tokenizer` to
        tokenize the input text. Defaults to a space tokenization.
      dtype: The embedding type.
    """
    super(CharConvEmbedder, self).__init__(
        vocabulary_file_key,
        embedding_size,
        dropout=dropout,
        tokenizer=tokenizer,
        dtype=dtype)
    self.output_size = num_outputs
    self.kernel_size = kernel_size
    self.stride = stride
    self.num_oov_buckets = 1

  def make_inputs(self, features, training=None):
    inputs = features["char_ids"]
    outputs = self._embed(inputs, training)

    # Merge batch and sequence timesteps dimensions.
    outputs = tf.reshape(outputs, [-1, tf.shape(inputs)[-1], self.embedding_size])

    # Pad on both sides.
    outputs = tf.pad(outputs, [[0, 0], [self.kernel_size - 1, self.kernel_size - 1], [0, 0]])
    outputs.set_shape((None, None, self.embedding_size))

    outputs = tf.layers.conv1d(
        outputs,
        self.output_size,
        self.kernel_size,
        strides=self.stride)

    # Max pooling over depth.
    outputs = tf.reduce_max(outputs, axis=1)

    # Split batch and sequence timesteps dimensions.
    outputs = tf.reshape(outputs, [-1, tf.shape(inputs)[1], self.output_size])

    return outputs


class CharRNNEmbedder(CharEmbedder):
  """An inputter that runs a single RNN layer over character embeddings."""

  def __init__(self,
               vocabulary_file_key,
               embedding_size,
               num_units,
               dropout=0.2,
               encoding="average",
               cell_class=None,
               tokenizer=None,
               dtype=tf.float32):
    """Initializes the parameters of the character RNN embedder.

    Args:
      vocabulary_file_key: The meta configuration key of the vocabulary file
        containing one character per line.
      embedding_size: The size of the character embedding.
      num_units: The number of units in the RNN layer.
      dropout: The probability to drop units in the embedding and the RNN
        outputs.
      encoding: "average" or "last" (case insensitive), the encoding vector to
        extract from the RNN outputs.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a LSTM cell.
      tokenizer: An optional :class:`opennmt.tokenizers.tokenizer.Tokenizer` to
        tokenize the input text. Defaults to a space tokenization.
      dtype: The embedding type.

    Raises:
      ValueError: if :obj:`encoding` is invalid.
    """
    super(CharRNNEmbedder, self).__init__(
        vocabulary_file_key,
        embedding_size,
        dropout=dropout,
        tokenizer=tokenizer,
        dtype=dtype)
    self.num_units = num_units
    self.cell_class = cell_class
    self.encoding = encoding.lower()
    if self.encoding not in ("average", "last"):
      raise ValueError("Invalid encoding vector: {}".format(self.encoding))

  def make_inputs(self, features, training=None):
    inputs = features["char_ids"]
    flat_inputs = tf.reshape(inputs, [-1, tf.shape(inputs)[-1]])
    embeddings = self._embed(flat_inputs, training)
    sequence_length = tf.count_nonzero(flat_inputs, axis=1)

    cell = build_cell(
        1,
        self.num_units,
        tf.estimator.ModeKeys.TRAIN if training else None,
        dropout=self.dropout,
        cell_class=self.cell_class)
    rnn_outputs, rnn_state = tf.nn.dynamic_rnn(
        cell,
        embeddings,
        sequence_length=sequence_length,
        dtype=embeddings.dtype)

    if self.encoding == "average":
      encoding = tf.reduce_mean(rnn_outputs, axis=1)
    elif self.encoding == "last":
      encoding = last_encoding_from_state(rnn_state)

    outputs = tf.reshape(encoding, [-1, tf.shape(inputs)[1], self.num_units])
    return outputs



"""Sequence classifier."""

import tensorflow as tf

from opennmt import inputters
from opennmt.models.model import Model
from opennmt.utils.cell import last_encoding_from_state
from opennmt.utils.misc import print_bytes
from opennmt.utils.losses import cross_entropy_loss


class SequenceClassifier(Model):
  """A sequence classifier."""

  def __init__(self,
               inputter,
               encoder,
               labels_vocabulary_file_key,
               encoding="average",
               daisy_chain_variables=False,
               name="seqclassifier"):
    """Initializes a sequence classifier.

    Args:
      inputter: A :class:`opennmt.inputters.inputter.Inputter` to process the
        input data.
      encoder: A :class:`opennmt.encoders.encoder.Encoder` to encode the input.
      labels_vocabulary_file_key: The data configuration key of the labels
        vocabulary file containing one label per line.
      encoding: "average" or "last" (case insensitive), the encoding vector to
        extract from the encoder outputs.
      daisy_chain_variables: If ``True``, copy variables in a daisy chain
        between devices for this model. Not compatible with RNN based models.
      name: The name of this model.

    Raises:
      ValueError: if :obj:`encoding` is invalid.
    """
    super(SequenceClassifier, self).__init__(
        name,
        features_inputter=inputter,
        labels_inputter=ClassInputter(labels_vocabulary_file_key),
        daisy_chain_variables=daisy_chain_variables)
    self.encoder = encoder
    self.encoding = encoding.lower()
    if self.encoding not in ("average", "last"):
      raise ValueError("Invalid encoding vector: {}".format(self.encoding))

  def _call(self, features, labels, params, mode):
    training = mode == tf.estimator.ModeKeys.TRAIN
    with tf.variable_scope("encoder"):
      inputs = self.features_inputter.make_inputs(features, training=training)
      encoder_outputs, encoder_state, _ = self.encoder.encode(
          inputs,
          sequence_length=self.features_inputter.get_length(features),
          mode=mode)

    if self.encoding == "average":
      encoding = tf.reduce_mean(encoder_outputs, axis=1)
    elif self.encoding == "last":
      encoding = last_encoding_from_state(encoder_state)

    with tf.variable_scope("generator"):
      logits = tf.layers.dense(
          encoding,
          self.labels_inputter.vocabulary_size)

    if mode != tf.estimator.ModeKeys.TRAIN:
      labels_vocab_rev = self.labels_inputter.vocabulary_lookup_reverse()
      classes_prob = tf.nn.softmax(logits)
      classes_id = tf.argmax(classes_prob, axis=1)
      predictions = {
          "classes": labels_vocab_rev.lookup(classes_id)
      }
    else:
      predictions = None

    return logits, predictions

  def compute_loss(self, outputs, labels, training=True, params=None):
    if params is None:
      params = {}
    return cross_entropy_loss(
        outputs,
        labels["classes_id"],
        label_smoothing=params.get("label_smoothing", 0.0),
        training=training)

  def compute_metrics(self, predictions, labels):
    return {
        "accuracy": tf.metrics.accuracy(labels["classes"], predictions["classes"])
    }

  def print_prediction(self, prediction, params=None, stream=None):
    print_bytes(prediction["classes"], stream=stream)


class ClassInputter(inputters.TextInputter):
  """Reading class from a text file."""

  def __init__(self, vocabulary_file_key):
    super(ClassInputter, self).__init__(
        vocabulary_file_key=vocabulary_file_key, num_oov_buckets=0)

  def make_features(self, element=None, features=None, training=None):
    return {
        "classes": element,
        "classes_id": self.vocabulary.lookup(element)
    }



"""Script to update vocabularies in a checkpoint."""

import argparse

import tensorflow as tf

from opennmt.utils import checkpoint


def main():
  tf.logging.set_verbosity(tf.logging.INFO)

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("--model_dir", required=True,
                      help="The model directory containing the checkpoints.")
  parser.add_argument("--output_dir", required=True,
                      help="The output directory where the updated checkpoint will be saved.")
  parser.add_argument("--src_vocab", required=True,
                      help="Path to the current source vocabulary.")
  parser.add_argument("--tgt_vocab", required=True,
                      help="Path to the current target vocabulary.")
  parser.add_argument("--new_src_vocab", default=None,
                      help="Path to the new source vocabulary.")
  parser.add_argument("--new_tgt_vocab", default=None,
                      help="Path to the new target vocabulary.")
  parser.add_argument("--mode", default="merge", choices=["merge", "replace"],
                      help="Vocabulary update mode.")
  args = parser.parse_args()
  checkpoint.update_vocab(args.model_dir,
                          args.output_dir,
                          args.src_vocab,
                          args.tgt_vocab,
                          new_src_vocab=args.new_src_vocab,
                          new_tgt_vocab=args.new_tgt_vocab,
                          mode=args.mode,
                          session_config=tf.ConfigProto(device_count={"GPU": 0}))


if __name__ == "__main__":
  main()



"""Define the Google's Transformer model."""

import tensorflow as tf

from opennmt.models.sequence_to_sequence import SequenceToSequence, EmbeddingsSharingLevel
from opennmt.encoders.encoder import ParallelEncoder
from opennmt.encoders.self_attention_encoder import SelfAttentionEncoder
from opennmt.decoders.self_attention_decoder import SelfAttentionDecoder
from opennmt.layers.position import SinusoidalPositionEncoder
from opennmt.utils.misc import merge_dict


class Transformer(SequenceToSequence):
  """Attention-based sequence-to-sequence model as described in
  https://arxiv.org/abs/1706.03762.
  """

  def __init__(self,
               source_inputter,
               target_inputter,
               num_layers,
               num_units,
               num_heads,
               ffn_inner_dim,
               dropout=0.1,
               attention_dropout=0.1,
               relu_dropout=0.1,
               position_encoder=SinusoidalPositionEncoder(),
               decoder_self_attention_type="scaled_dot",
               share_embeddings=EmbeddingsSharingLevel.NONE,
               share_encoders=False,
               alignment_file_key="train_alignments",
               name="transformer"):
    """Initializes a Transformer model.

    Args:
      source_inputter: A :class:`opennmt.inputters.inputter.Inputter` to process
        the source data. If this inputter returns parallel inputs, a multi
        source Transformer architecture will be constructed.
      target_inputter: A :class:`opennmt.inputters.inputter.Inputter` to process
        the target data. Currently, only the
        :class:`opennmt.inputters.text_inputter.WordEmbedder` is supported.
      num_layers: The shared number of layers.
      num_units: The number of hidden units.
      num_heads: The number of heads in each self-attention layers.
      ffn_inner_dim: The inner dimension of the feed forward layers.
      dropout: The probability to drop units in each layer output.
      attention_dropout: The probability to drop units from the attention.
      relu_dropout: The probability to drop units from the ReLU activation in
        the feed forward layer.
      position_encoder: A :class:`opennmt.layers.position.PositionEncoder` to
        apply on the inputs.
      decoder_self_attention_type: Type of self attention in the decoder,
        "scaled_dot" or "average" (case insensitive).
      share_embeddings: Level of embeddings sharing, see
        :class:`opennmt.models.sequence_to_sequence.EmbeddingsSharingLevel`
        for possible values.
      share_encoders: In case of multi source architecture, whether to share the
        separate encoders parameters or not.
      alignment_file_key: The data configuration key of the training alignment
        file to support guided alignment.
      name: The name of this model.
    """
    encoders = [
        SelfAttentionEncoder(
            num_layers,
            num_units=num_units,
            num_heads=num_heads,
            ffn_inner_dim=ffn_inner_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            relu_dropout=relu_dropout,
            position_encoder=position_encoder)
        for _ in range(source_inputter.num_outputs)]
    if len(encoders) > 1:
      encoder = ParallelEncoder(
          encoders,
          outputs_reducer=None,
          states_reducer=None,
          share_parameters=share_encoders)
    else:
      encoder = encoders[0]
    decoder = SelfAttentionDecoder(
        num_layers,
        num_units=num_units,
        num_heads=num_heads,
        ffn_inner_dim=ffn_inner_dim,
        dropout=dropout,
        attention_dropout=attention_dropout,
        relu_dropout=relu_dropout,
        position_encoder=position_encoder,
        self_attention_type=decoder_self_attention_type)

    self._num_units = num_units
    super(Transformer, self).__init__(
        source_inputter,
        target_inputter,
        encoder,
        decoder,
        share_embeddings=share_embeddings,
        alignment_file_key=alignment_file_key,
        daisy_chain_variables=True,
        name=name)

  def auto_config(self, num_devices=1):
    config = super(Transformer, self).auto_config(num_devices=num_devices)
    return merge_dict(config, {
        "params": {
            "average_loss_in_time": True,
            "label_smoothing": 0.1,
            "optimizer": "LazyAdamOptimizer",
            "optimizer_params": {
                "beta1": 0.9,
                "beta2": 0.998
            },
            "learning_rate": 2.0,
            "decay_type": "noam_decay_v2",
            "decay_params": {
                "model_dim": self._num_units,
                "warmup_steps": 8000
            }
        },
        "train": {
            "effective_batch_size": 25000,
            "batch_size": 3072,
            "batch_type": "tokens",
            "maximum_features_length": 100,
            "maximum_labels_length": 100,
            "keep_checkpoint_max": 8,
            "average_last_checkpoints": 8
        }
    })

  def _initializer(self, params):
    return tf.variance_scaling_initializer(
        mode="fan_avg", distribution="uniform", dtype=self.dtype)



"""Utilities to run execution in parallel."""

import six

import tensorflow as tf

from tensorflow.python.client import device_lib
from tensorflow.python.estimator.util import fn_args


class GraphDispatcher(object):
  """Helper class to replicate graph parts on multiple devices and dispatch
  sharded batches.
  """

  def __init__(self,
               num_devices=None,
               daisy_chain_variables=True,
               devices=None,
               session_config=None):
    """Initializes the dispatcher.

    Args:
      num_devices: The number of devices to dispatch on.
      daisy_chain_variables: If ``True``, variables are copied in a daisy chain
        fashion between devices (credits to Tensor2Tensor).
      devices: List of devices to use (takes priority over :obj:`num_devices`).
      session_config: Session configuration to use when querying available
        devices.

    Raises:
      ValueError: if the number of visible devices is lower than
        :obj:`num_devices`.
    """
    if devices:
      self._devices = devices
    elif num_devices is not None:
      self._devices = get_devices(num_devices=num_devices, session_config=session_config)
    else:
      self._devices = [None]
    self._n = len(self._devices)
    self._daisy_chain_variables = daisy_chain_variables

  def shard(self, data):
    """Shards a structure of ``tf.Tensor`` for dispatching.

    Args:
      data: A ``tf.Tensor`` of dictionary of ``tf.Tensor``.

    Returns:
      A list of the same ``tf.Tensor`` structure.
    """
    return split_batch(data, self._n)

  def repeat(self, data):
    """Ensures that the object is dispatchable list.

    Args:
      data: The object to convert.

    Returns:
      :obj:`data` if it is valid list or a list where :obj:`data` is replicated.

    Raises:
      ValueError: if :obj:`data` is a non dispatchable list.
    """
    if isinstance(data, list):
      if len(data) != self._n:
        raise ValueError("List arguments must contain %d elements, saw %d instead"
                         % (self._n, len(data)))
      return data
    else:
      return [data] * self._n

  def _parallel_args(self, *args, **kwargs):
    """Makes each argument dispatchable."""
    if args:
      parallel_args = [self.repeat(arg) for arg in args]
      parallel_args = [list(arg) for arg in zip(*parallel_args)]
    else:
      parallel_args = [[] for _ in range(self._n)]
    parallel_kwargs = [{} for _ in range(self._n)]
    for k, v in six.iteritems(kwargs):
      values = self.repeat(v)
      for i in range(self._n):
        parallel_kwargs[i][k] = values[i]
    return parallel_args, parallel_kwargs

  def __call__(self, fun, *args, **kwargs):
    """Dispatches :obj:`fun` calls accross devices.

    Each argument must either not be a list or a list with length the number of
    devices used for dispatching.

    Args:
      fun: A callable.
      *args: The callable arguments.
      **kwargs: The callable keyword arguments.

    Returns:
      The sharded outputs of :obj:`fun`.
    """
    funs = self.repeat(fun)
    args, kwargs = self._parallel_args(*args, **kwargs)

    outputs = []
    cache = {}
    tensor_to_var = {}

    for i, device in enumerate(self._devices):

      # pylint: disable=cell-var-from-loop
      def _daisy_chain_getter(getter, name, *args, **kwargs):
        """Get a variable and cache in a daisy chain."""
        # Copyright 2017 The Tensor2Tensor Authors.
        # Licensed under the Apache License, Version 2.0
        device_var_key = (device, name)
        if device_var_key in cache:
          # if we have the variable on the correct device, return it.
          return cache[device_var_key]
        if name in cache:
          # if we have it on a different device, copy it from the last device
          last_device_v = cache[name]
          var = tensor_to_var[last_device_v]
          v = tf.identity(last_device_v)
        else:
          var = getter(name, *args, **kwargs)
          v = tf.identity(var._ref())  # pylint: disable=protected-access

        # keep track of the original variable
        tensor_to_var[v] = var
        v.read_value = lambda: tf.identity(v)
        v.assign_sub = var.assign_sub
        # update the cache
        cache[name] = v
        cache[device_var_key] = v
        return v

      custom_getter = None
      if self._daisy_chain_variables:
        custom_getter = _daisy_chain_getter

      with tf.variable_scope(
          tf.get_variable_scope(),
          reuse=True if i > 0 else None,
          custom_getter=custom_getter):
        if device is None:
          outputs.append(funs[i](*args[i], **kwargs[i]))
        else:
          with tf.device(device):
            outputs.append(funs[i](*args[i], **kwargs[i]))

    # If the function returned a tuple, also return a tuple of sharded results.
    if isinstance(outputs[0], tuple):
      outputs = tuple(list(output) for output in zip(*outputs))

    return outputs


def split_batch(data, num_shards):
  """Split data into shards."""

  def _split_dictionary(dictionary):
    """Split a dictionary into shards."""
    shards = [{} for _ in range(num_shards)]
    for name, tensor in six.iteritems(dictionary):
      if isinstance(tensor, tf.SparseTensor):
        for i, shard in enumerate(tf.sparse_split(sp_input=tensor, num_split=num_shards, axis=0)):
          shards[i][name] = shard
      else:
        for i, shard in enumerate(tf.split(tensor, num_shards)):
          shards[i][name] = shard
    return shards

  with tf.name_scope("split_inputs"):
    if data is None:
      data_shards = None
    elif isinstance(data, dict):
      data_shards = _split_dictionary(data)
    else:
      data_shards = tf.split(data, num_shards)

  return data_shards

def get_devices(num_devices=None, session_config=None):
  """Returns available devices.

  Args:
    num_devices: The number of devices to get.
    session_config: An optional session configuration to use when querying
      available devices.

  Returns:
    A list of devices.

  Raises:
    ValueError: if :obj:`num_devices` is set but the number of visible devices
      is lower than it.
  """
  kwargs = {}
  if "session_config" in fn_args(device_lib.list_local_devices):
    kwargs["session_config"] = session_config
  else:
    # Create a first session to enforce config, otherwise list_local_devices()
    # will run some initialization with default options.
    _ = tf.Session(config=session_config)
  devices = [x.name for x in device_lib.list_local_devices(**kwargs) if x.device_type == "GPU"]
  if not devices:
    return [None]
  elif num_devices is None:
    return devices
  elif len(devices) < num_devices:
    raise ValueError("Only %d devices are visible but %d were requested"
                     % (len(devices), num_devices))
  return devices[:num_devices]



# -*- coding: utf-8 -*-

"""Define base tokenizers."""

import sys
import os
import abc
import copy
import six

import tensorflow as tf
import yaml

from opennmt.utils.misc import print_bytes
from opennmt.utils import compat


def _make_config_asset_file(config, asset_path):
  asset_config = copy.deepcopy(config)
  for key, value in six.iteritems(asset_config):
    # Only keep the basename for files (that should also be registered as assets).
    if isinstance(value, six.string_types) and compat.gfile_exists(value):
      asset_config[key] = os.path.basename(value)
  with open(asset_path, "w") as asset_file:
    yaml.dump(asset_config, stream=asset_file, default_flow_style=False)


@six.add_metaclass(abc.ABCMeta)
class Tokenizer(object):
  """Base class for tokenizers."""

  def __init__(self, configuration_file_or_key=None, params=None):
    """Initializes the tokenizer.

    Args:
      configuration_file_or_key: The YAML configuration file or a the key to
        the YAML configuration file.
    """
    self._configuration_key = None
    if params is not None:
      self._config = params
    else:
      self._config = {}
      if configuration_file_or_key is not None and compat.gfile_exists(configuration_file_or_key):
        configuration_file = configuration_file_or_key
        with compat.gfile_open(configuration_file, mode="rb") as conf_file:
          self._config = yaml.load(conf_file)
      else:
        self._configuration_key = configuration_file_or_key

  def initialize(self, metadata, asset_dir=None, asset_prefix=""):
    """Initializes the tokenizer (e.g. load BPE models).

    Args:
      metadata: A dictionary containing additional metadata set
        by the user.
      asset_dir: The directory where assets can be written. If ``None``, no
        assets are returned.
      asset_prefix: The prefix to attach to assets filename.

    Returns:
      A dictionary containing additional assets used by the tokenizer.
    """
    if self._configuration_key is not None:
      configuration = metadata[self._configuration_key]
      if isinstance(configuration, dict):
        self._config = configuration
      else:
        with compat.gfile_open(configuration, mode="rb") as conf_file:
          self._config = yaml.load(conf_file)
    if asset_dir is not None:
      return self.export_assets(asset_dir, asset_prefix=asset_prefix)
    return {}

  def export_assets(self, asset_dir, asset_prefix=""):
    """Exports assets for this tokenizer.

    Args:
      asset_dir: The directory where assets can be written.
      asset_prefix: The prefix to attach to assets filename.

    Returns:
      A dictionary containing additional assets used by the tokenizer.
    """
    assets = {}
    if self._config:
      asset_name = "%stokenizer_config.yml" % asset_prefix
      asset_path = os.path.join(asset_dir, asset_name)
      _make_config_asset_file(self._config, asset_path)
      assets[asset_name] = asset_path
    return assets

  def tokenize_stream(self, input_stream=sys.stdin, output_stream=sys.stdout, delimiter=" "):
    """Tokenizes a stream of sentences.

    Args:
      input_stream: The input stream.
      output_stream: The output stream.
      delimiter: The token delimiter to use for text serialization.
    """
    for line in input_stream:
      line = line.strip()
      tokens = self.tokenize(line)
      merged_tokens = delimiter.join(tokens)
      print_bytes(tf.compat.as_bytes(merged_tokens), stream=output_stream)

  def detokenize_stream(self, input_stream=sys.stdin, output_stream=sys.stdout, delimiter=" "):
    """Detokenizes a stream of sentences.

    Args:
      input_stream: The input stream.
      output_stream: The output stream.
      delimiter: The token delimiter used for text serialization.
    """
    for line in input_stream:
      tokens = line.strip().split(delimiter)
      string = self.detokenize(tokens)
      print_bytes(tf.compat.as_bytes(string), stream=output_stream)

  def tokenize(self, text):
    """Tokenizes text.

    Args:
      text: The text to tokenize as a ``tf.Tensor`` or Python string.

    Returns:
      A 1-D string ``tf.Tensor`` if :obj:`text` is a ``tf.Tensor`` or a list of
      Python unicode strings otherwise.

    Raises:
      ValueError: if the rank of :obj:`text` is greater than 0.
    """
    if compat.is_tensor(text):
      rank = len(text.get_shape().as_list())
      if rank == 0:
        return self._tokenize_tensor(text)
      else:
        raise ValueError("Unsupported tensor rank for tokenization: {}".format(rank))
    else:
      text = tf.compat.as_text(text)
      return self._tokenize_string(text)

  def detokenize(self, tokens, sequence_length=None):
    """Detokenizes tokens.

    The Tensor version supports batches of tokens.

    Args:
      tokens: The tokens as a 1-D or 2-D ``tf.Tensor`` or list of Python
        strings.
      sequence_length: The length of each sequence. Required if :obj:`tokens`
        is a ``tf.Tensor``.

    Returns:
      A 0-D or 1-D string ``tf.Tensor`` if :obj:`tokens` is a ``tf.Tensor`` or a
      Python unicode strings otherwise.

    Raises:
      ValueError: if the rank of :obj:`tokens` is greater than 2.
      ValueError: if :obj:`tokens` is a 2-D ``tf.Tensor`` and
        :obj:`sequence_length` is not set.
    """
    if compat.is_tensor(tokens):
      rank = len(tokens.get_shape().as_list())
      if rank == 1:
        return self._detokenize_tensor(tokens)
      elif rank == 2:
        if sequence_length is None:
          raise ValueError("sequence_length is required for Tensor detokenization")
        return self._detokenize_batch_tensor(tokens, sequence_length)
      else:
        raise ValueError("Unsupported tensor rank for detokenization: {}".format(rank))
    else:
      tokens = [tf.compat.as_text(token) for token in tokens]
      return self._detokenize_string(tokens)

  def _tokenize_tensor(self, text):
    """Tokenizes a tensor.

    When not overriden, this default implementation calls the string-based
    tokenization.

    Args:
      text: A 1-D string ``tf.Tensor``.

    Returns:
      A 1-D string ``tf.Tensor``.
    """
    if compat.tf_supports("py_function"):
      def _python_wrapper(string_t):
        string = tf.compat.as_text(string_t.numpy())
        tokens = self._tokenize_string(string)
        return tf.constant(tokens)
      tokens = tf.py_function(_python_wrapper, [text], tf.string)
      tokens.set_shape([None])
      return tokens

    text = tf.py_func(
        lambda x: tf.compat.as_bytes("\0".join(self.tokenize(x))), [text], tf.string)
    tokens = tf.string_split([text], delimiter="\0").values
    return tokens

  def _detokenize_tensor(self, tokens):
    """Detokenizes tokens.

    When not overriden, this default implementation calls the string-based
    detokenization.

    Args:
      tokens: A 1-D ``tf.Tensor``.

    Returns:
      A 0-D string ``tf.Tensor``.
    """
    if compat.tf_supports("py_function"):
      def _python_wrapper(tokens_t):
        tokens = [tf.compat.as_text(s) for s in tokens_t.numpy()]
        string = self._detokenize_string(tokens)
        return tf.constant(string)
      text = tf.py_function(_python_wrapper, [tokens], tf.string)
      text.set_shape([])
      return text
    return tf.py_func(self.detokenize, [tokens], tf.string)

  def _detokenize_batch_tensor(self, tokens, sequence_length):
    """Detokenizes a batch of tokens.

    When not overriden, this default implementation calls _detokenize_tensor on
    each tensor within the batch.

    Args:
      tokens: A 2-D ``tf.Tensor``.

    Returns:
      A 1-D string ``tf.Tensor``.
    """
    return tf.map_fn(
        lambda x: self._detokenize_tensor(x[0][:x[1]]),
        (tokens, sequence_length),
        dtype=tf.string,
        back_prop=False)

  @abc.abstractmethod
  def _tokenize_string(self, text):
    """Tokenizes a Python unicode string.

    This method should be thread-safe.

    Args:
      text: A Python unicode string.

    Returns:
      A list of Python unicode strings.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def _detokenize_string(self, tokens):
    """Detokenizes tokens.

    Args:
      tokens: A list of Python unicode strings.

    Returns:
      A unicode Python string.
    """
    raise NotImplementedError()


class SpaceTokenizer(Tokenizer):
  """A tokenizer that splits on spaces."""

  def _tokenize_tensor(self, text):
    if compat.tf_supports("string.splits"):
      return tf.strings.split([text]).values
    else:
      return tf.string_split([text], delimiter=" ").values

  def _detokenize_tensor(self, tokens):
    reduce_join = compat.tf_compat(v2="strings.reduce_join", v1="reduce_join")
    return reduce_join(tokens, axis=0, separator=" ")

  def _tokenize_string(self, text):
    return text.split()

  def _detokenize_string(self, tokens):
    return " ".join(tokens)


class CharacterTokenizer(Tokenizer):
  """A tokenizer that splits unicode characters."""

  def _tokenize_tensor(self, text):
    if compat.tf_supports("strings.unicode_split"):
      text = tf.strings.regex_replace(text, " ", "▁")
      return tf.strings.unicode_split(text, "UTF-8")
    else:
      return super(CharacterTokenizer, self)._tokenize_tensor(text)

  def _detokenize_tensor(self, tokens):
    if compat.tf_supports("strings.reduce_join"):
      text = tf.strings.reduce_join(tokens, axis=0)
      return tf.strings.regex_replace(text, "▁", " ")
    else:
      return super(CharacterTokenizer, self)._detokenize_tensor(tokens)

  def _tokenize_string(self, text):
    return list(text.replace(" ", u"▁"))

  def _detokenize_string(self, tokens):
    return "".join(tokens).replace(u"▁", " ")



"""Base class for models."""

from __future__ import print_function

import abc
import six

import tensorflow as tf

from opennmt import estimator
from opennmt import inputters
from opennmt.utils import compat
from opennmt.utils.optim import optimize_loss


@six.add_metaclass(abc.ABCMeta)
class Model(tf.keras.layers.Layer):
  """Base class for models."""

  def __init__(self,
               name,
               features_inputter=None,
               labels_inputter=None,
               daisy_chain_variables=False,
               dtype=None,
               examples_inputter=None):
    if examples_inputter is None:
      examples_inputter = inputters.ExampleInputter(features_inputter, labels_inputter)
    self.examples_inputter = examples_inputter
    if dtype is None:
      dtype = self.features_inputter.dtype
    super(Model, self).__init__(name=name, dtype=dtype)
    self.daisy_chain_variables = daisy_chain_variables

  @property
  def unsupervised(self):
    """Unsupervised model."""
    return not hasattr(self.examples_inputter, "labels_inputter")

  @property
  def features_inputter(self):
    """The inputter producing features."""
    return getattr(self.examples_inputter, "features_inputter", self.examples_inputter)

  @property
  def labels_inputter(self):
    """The inputter producing labels."""
    return self.examples_inputter.labels_inputter

  def auto_config(self, num_devices=1):
    """Returns automatic configuration values specific to this model.

    Args:
      num_devices: The number of devices used for the training.

    Returns:
      A partial training configuration.
    """
    _ = num_devices
    return {}

  def initialize(self, metadata):
    """Initializes the model from the data configuration.

    Args:
      metadata: A dictionary containing additional data configuration set
        by the user (e.g. vocabularies, tokenization, pretrained embeddings,
        etc.).
    """
    self.examples_inputter.initialize(metadata)

  def __call__(self, features, labels, params, mode, config=None):  # pylint: disable=arguments-differ
    """Calls the model function.

    Returns:
      outputs: The model outputs (usually unscaled probabilities).
        Optional if :obj:`mode` is ``tf.estimator.ModeKeys.PREDICT``.
      predictions: The model predictions.
        Optional if :obj:`mode` is ``tf.estimator.ModeKeys.TRAIN``.

    See Also:
      ``tf.estimator.Estimator`` 's ``model_fn`` argument for more details about
      the arguments of this function.
    """
    with tf.variable_scope(self.name, initializer=self._initializer(params)):
      if not compat.reuse():
        self._build()  # Always rebuild unless the scope is marked for reuse.
      return self._call(features, labels, params, mode)

  def _initializer(self, params):
    """Returns the global initializer for this model.

    Args:
      params: A dictionary of hyperparameters.

    Returns:
      The initializer.
    """
    param_init = params.get("param_init")
    if param_init is not None:
      return tf.random_uniform_initializer(
          minval=-param_init, maxval=param_init, dtype=self.dtype)
    return None

  def _build(self):
    """Builds stateful layers."""
    return

  @abc.abstractmethod
  def _call(self, features, labels, params, mode):
    """Creates the graph.

    Returns:
      outputs: The model outputs (usually unscaled probabilities).
        Optional if :obj:`mode` is ``tf.estimator.ModeKeys.PREDICT``.
      predictions: The model predictions.
        Optional if :obj:`mode` is ``tf.estimator.ModeKeys.TRAIN``.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def compute_loss(self, outputs, labels, training=True, params=None):
    """Computes the loss.

    Args:
      outputs: The model outputs (usually unscaled probabilities).
      labels: The dict of labels ``tf.Tensor``.
      training: Compute training loss.
      params: A dictionary of hyperparameters.

    Returns:
      The loss or a tuple containing the computed loss and the loss to display.
    """
    raise NotImplementedError()

  def optimize_loss(self, loss, params=None, hvd=None):
    """Returns the loss optimization op.

    Args:
      loss: The loss to optimize.
      params: A dictionary of hyperparameters.
      hvd: Optional Horovod object.

    Returns:
      The training op and optionally a list of extra variables to initialize.
    """
    if params is None:
      params = {}
    mixed_precision = self.dtype == tf.float16
    return optimize_loss(loss, params, mixed_precision=mixed_precision, hvd=hvd)

  def compute_metrics(self, predictions, labels):  # pylint: disable=unused-argument
    """Computes additional metrics on the predictions.

    Args:
      predictions: The model predictions.
      labels: The dict of labels ``tf.Tensor``.

    Returns:
      A dict of metrics. See the ``eval_metric_ops`` field of
      ``tf.estimator.EstimatorSpec``.
    """
    return None

  def input_fn(self,
               mode,
               batch_size,
               metadata=None,
               features_file=None,
               labels_file=None,
               batch_type="examples",
               batch_multiplier=1,
               bucket_width=None,
               single_pass=False,
               num_threads=None,
               sample_buffer_size=None,
               prefetch_buffer_size=None,
               maximum_features_length=None,
               maximum_labels_length=None,
               num_shards=1,
               shard_index=0):
    """Returns an input function.

    Args:
      mode: A ``tf.estimator.ModeKeys`` mode.
      batch_size: The batch size to use.
      metadata: A dictionary containing additional metadata set
        by the user. Required if ``Model.initialize()`` has not been called.
      features_file: The file containing input features.
      labels_file: The file containing output labels.
      batch_type: The training batching stragety to use: can be "examples" or
        "tokens".
      batch_multiplier: The batch size multiplier to prepare splitting accross
         replicated graph parts.
      bucket_width: The width of the length buckets to select batch candidates
        from. ``None`` to not constrain batch formation.
      single_pass: If ``True``, makes a single pass over the training data.
      num_threads: The number of elements processed in parallel.
      sample_buffer_size: The number of elements from which to sample.
      prefetch_buffer_size: The number of batches to prefetch asynchronously. If
        ``None``, use an automatically tuned value on TensorFlow 1.8+ and 1 on
        older versions.
      maximum_features_length: The maximum length or list of maximum lengths of
        the features sequence(s). ``None`` to not constrain the length.
      maximum_labels_length: The maximum length of the labels sequence.
        ``None`` to not constrain the length.
      num_shards: The number of data shards (usually the number of workers in a
        distributed setting).
      shard_index: The shard index this input pipeline should read from.

    Returns:
      A callable that returns the next element.

    See Also:
      ``tf.estimator.Estimator``.
    """
    if metadata is not None:
      self.initialize(metadata)
    return estimator.make_input_fn(
        self,
        mode,
        batch_size,
        features_file,
        labels_file=labels_file,
        batch_type=batch_type,
        batch_multiplier=batch_multiplier,
        bucket_width=bucket_width,
        maximum_features_length=maximum_features_length,
        maximum_labels_length=maximum_labels_length,
        shuffle_buffer_size=sample_buffer_size,
        single_pass=single_pass,
        num_shards=num_shards,
        shard_index=shard_index,
        num_threads=num_threads,
        prefetch_buffer_size=prefetch_buffer_size,
        return_dataset=False)

  def model_fn(self, num_devices=1, eval_prediction_hooks_fn=None, devices=None, hvd=None):
    """Returns the model function.

    Args:
      num_devices: The number of devices used for training.
      eval_prediction_hooks_fn: A callable that takes the model predictions
        during evaluation and return an iterable of evaluation hooks (e.g. for
        saving predictions on disk, running external evaluators, etc.).
      devices: The list of devices used for training, if known.
      hvd: Optional Horovod object.

    See Also:
      ``tf.estimator.Estimator`` 's ``model_fn`` argument for more details about
      arguments and the returned value.
    """
    return estimator.make_model_fn(
        self,
        eval_prediction_hooks_fn=eval_prediction_hooks_fn,
        num_devices=num_devices,
        devices=devices,
        hvd=hvd)

  def serving_input_fn(self, metadata=None):
    """Returns the serving input function.

    Args:
      metadata: A dictionary containing additional metadata set
        by the user. Required if ``Model.initialize()`` has not been called.

    Returns:
      A callable that returns a ``tf.estimator.export.ServingInputReceiver``.
    """
    if metadata is not None:
      self.initialize(metadata)
    return estimator.make_serving_input_fn(self)

  def get_assets(self, metadata=None, asset_dir=None):
    """Returns additional assets used by this model.

    Args:
      metadata: A dictionary containing additional metadata set
        by the user. Required if ``Model.initialize()`` has not been called.
      asset_dir: The directory where assets can be written.

    Returns:
      A dictionary of additional assets.
    """
    if metadata is not None:
      self.initialize(metadata)
    return self.examples_inputter.export_assets(asset_dir)

  def print_prediction(self, prediction, params=None, stream=None):
    """Prints the model prediction.

    Args:
      prediction: The evaluated prediction.
      params: (optional) Dictionary of formatting parameters.
      stream: (optional) The stream to print to.
    """
    _ = params
    print(prediction, file=stream)



"""Checkpoint utilities."""

import os
import six

import tensorflow as tf
import numpy as np

from opennmt.utils.vocab import Vocab
from opennmt.utils.misc import count_lines


def _get_vocabulary_mapping(current_vocab_path, new_vocab_path, mode):
  """Maps vocabulary new indices to old ones. -1 means that the entry is new."""
  current_vocab = Vocab(from_file=current_vocab_path)
  new_vocab = Vocab(from_file=new_vocab_path)
  mapping = []
  if mode == "merge":
    final_vocab = Vocab(from_file=current_vocab_path)
    mapping = [i for i in range(current_vocab.size)]
    for new_word in new_vocab.words:
      if current_vocab.lookup(new_word) is None:
        mapping.append(-1)
        final_vocab.add(new_word)
  elif mode == "replace":
    final_vocab = new_vocab
    for new_word in new_vocab.words:
      idx = current_vocab.lookup(new_word)
      if idx is not None:
        mapping.append(idx)
      else:
        mapping.append(-1)
  mapping.append(current_vocab.size)  # <unk> token is always the last entry.
  return mapping, final_vocab

def _update_vocabulary_variable(variable, vocab_size, mapping):
  """Creates a new variable, possibly copying previous entries based on mapping."""
  dim = variable.shape.index(vocab_size)
  # Make the dimension to index the first.
  perm = list(range(len(variable.shape)))
  perm[0], perm[dim] = perm[dim], perm[0]
  variable_t = np.transpose(variable, axes=perm)
  new_shape = list(variable_t.shape)
  new_shape[0] = len(mapping)
  new_variable_t = np.zeros(new_shape, dtype=variable.dtype)
  for i, j in enumerate(mapping):
    if j >= 0:
      new_variable_t[i] = variable_t[j]
  new_variable = np.transpose(new_variable_t, axes=perm)
  return new_variable

def _update_vocabulary_variables(output_dir,
                                 variables,
                                 current_vocab_path,
                                 new_vocab_path,
                                 scope_name,
                                 mode):
  current_size = count_lines(current_vocab_path) + 1
  mapping, vocab = _get_vocabulary_mapping(current_vocab_path, new_vocab_path, mode)
  if mode == "merge":
    vocab_name = "%s-%s-%s" % (scope_name, mode, os.path.basename(new_vocab_path))
    vocab.serialize(os.path.join(output_dir, vocab_name))
  for name, tensor in six.iteritems(variables):
    if scope_name in name and any(d == current_size for d in tensor.shape):
      tf.logging.debug("Updating variable %s" % name)
      variables[name] = _update_vocabulary_variable(tensor, current_size, mapping)

def _variable_is_trainable(name, value):
  _ = name
  return value.dtype not in (np.int32, np.int64)  # Assume that int variables are not trainable.

def _create_checkpoint_from_variables(variables, output_dir, latest_step=None, session_config=None):
  # The straightforward approach would be to create new variables using a
  # constant_initializer. However, this would save the constant value in the
  # checkpoint meta file which would increase its size dramatically. Instead, we
  # create variables with their default initializer but run an assignment op
  # that writes the new value. Inspired by:
  # github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/bin/t2t_avg_all.py
  if "global_step" in variables:
    latest_step = variables["global_step"]
    del variables["global_step"]
  tf_vars = [
      tf.get_variable(
          name,
          shape=value.shape,
          dtype=tf.as_dtype(value.dtype),
          trainable=_variable_is_trainable(name, value))
      for name, value in six.iteritems(variables)]
  placeholders = [tf.placeholder(v.dtype, shape=v.shape) for v in tf_vars]
  assign_ops = [tf.assign(v, p) for (v, p) in zip(tf_vars, placeholders)]

  out_base_file = os.path.join(output_dir, "model.ckpt")
  global_step = tf.get_variable(
      "global_step",
      initializer=tf.constant(latest_step, dtype=tf.int64),
      trainable=False)
  saver = tf.train.Saver(tf.global_variables(), save_relative_paths=True)

  with tf.Session(config=session_config) as sess:
    sess.run(tf.global_variables_initializer())
    for p, assign_op, value in zip(placeholders, assign_ops, six.itervalues(variables)):
      sess.run(assign_op, {p: value})
    tf.logging.info("Saving new checkpoint to %s" % output_dir)
    saver.save(sess, out_base_file, global_step=global_step)

  return output_dir

def get_checkpoint_variables(checkpoint_path):
  """Returns variables included in a checkpoint.

  Args:
    checkpoint_path: Path to the checkpoint.

  Returns:
    A dictionary mapping variables name to value.
  """
  reader = tf.train.load_checkpoint(checkpoint_path)
  return {
      name:reader.get_tensor(name)
      for name in six.iterkeys(reader.get_variable_to_shape_map())}

def convert_checkpoint(checkpoint_path,
                       output_dir,
                       source_dtype,
                       target_type,
                       session_config=None):
  """Converts checkpoint variables from one dtype to another.

  Args:
    checkpoint_path: The path to the checkpoint to convert.
    output_dir: The directory that will contain the converted checkpoint.
    source_dtype: The data type to convert from.
    target_dtype: The data type to convert to.
    session_config: Optional configuration to use when creating the session.

  Returns:
    The path to the directory containing the converted checkpoint.

  Raises:
    ValueError: if :obj:`output_dir` points to the same directory as
      :obj:`checkpoint_path`.
  """
  if os.path.dirname(checkpoint_path) == output_dir:
    raise ValueError("Checkpoint and output directory must be different")
  variables = get_checkpoint_variables(checkpoint_path)
  for name, value in six.iteritems(variables):
    if not name.startswith("optim") and tf.as_dtype(value.dtype) == source_dtype:
      variables[name] = value.astype(target_type.as_numpy_dtype())
  return _create_checkpoint_from_variables(
      variables,
      output_dir,
      session_config=session_config)

def update_vocab(model_dir,
                 output_dir,
                 current_src_vocab,
                 current_tgt_vocab,
                 new_src_vocab=None,
                 new_tgt_vocab=None,
                 mode="merge",
                 session_config=None):
  """Updates the last checkpoint to support new vocabularies.

  This allows to add new words to a model while keeping the previously learned
  weights.

  Args:
    model_dir: The directory containing checkpoints (the most recent will be loaded).
    output_dir: The directory that will contain the converted checkpoint. In
      ``merge`` mode, updated vocabularies will also be saved in the directory.
    current_src_vocab: Path to the source vocabulary currently use in the model.
    current_tgt_vocab: Path to the target vocabulary currently use in the model.
    new_src_vocab: Path to the new source vocabulary to support.
    new_tgt_vocab: Path to the new target vocabulary to support.
    mode: Update mode: "merge" keeps all existing words and adds new words,
      "replace" makes the new vocabulary file the active one. In all modes,
      if an existing word appears in the new vocabulary, its learned weights
      are kept in the converted checkpoint.
    session_config: Optional configuration to use when creating the session.

  Returns:
    The path to the directory containing the converted checkpoint.

  Raises:
    ValueError: if :obj:`output_dir` is the same as :obj:`model_dir`, if
      :obj:`mode` is invalid, or if no checkpoints are found in
      :obj:`model_dir`.
  """
  if model_dir == output_dir:
    raise ValueError("Model and output directory must be different")
  if mode not in ("merge", "replace"):
    raise ValueError("invalid vocab update mode: %s" % mode)
  if new_src_vocab is None and new_tgt_vocab is None:
    return model_dir
  checkpoint_path = tf.train.latest_checkpoint(model_dir)
  if checkpoint_path is None:
    raise ValueError("No checkpoint found in directory %s" % model_dir)
  tf.logging.info("Updating vocabulary related variables in checkpoint %s" % checkpoint_path)
  variable_value = get_checkpoint_variables(checkpoint_path)
  if not tf.gfile.Exists(output_dir):
    tf.gfile.MakeDirs(output_dir)
  if new_src_vocab is not None:
    _update_vocabulary_variables(
        output_dir, variable_value, current_src_vocab, new_src_vocab, "encoder", mode)
  if new_tgt_vocab is not None:
    _update_vocabulary_variables(
        output_dir, variable_value, current_tgt_vocab, new_tgt_vocab, "decoder", mode)
  return _create_checkpoint_from_variables(
      variable_value,
      output_dir,
      session_config=session_config)


def average_checkpoints(model_dir, output_dir, max_count=8, session_config=None):
  """Averages checkpoints.

  Args:
    model_dir: The directory containing checkpoints.
    output_dir: The directory that will contain the averaged checkpoint.
    max_count: The maximum number of checkpoints to average.
    session_config: Configuration to use when creating the session.

  Returns:
    The path to the directory containing the averaged checkpoint.

  Raises:
    ValueError: if :obj:`output_dir` is the same as :obj:`model_dir`.
  """
  if model_dir == output_dir:
    raise ValueError("Model and output directory must be different")

  checkpoints_path = tf.train.get_checkpoint_state(model_dir).all_model_checkpoint_paths
  if len(checkpoints_path) > max_count:
    checkpoints_path = checkpoints_path[-max_count:]
  num_checkpoints = len(checkpoints_path)

  tf.logging.info("Averaging %d checkpoints..." % num_checkpoints)
  tf.logging.info("Listing variables...")

  new_variables = {}
  for i, checkpoint_path in enumerate(checkpoints_path):
    tf.logging.info("Loading checkpoint %s" % checkpoint_path)
    variables = get_checkpoint_variables(checkpoint_path)
    for name, value in six.iteritems(variables):
      if _variable_is_trainable(name, value):
        scaled_value = value / num_checkpoints
        if name in new_variables:
          new_variables[name] += scaled_value
        else:
          new_variables[name] = scaled_value
      elif i + 1 == num_checkpoints:  # Take non trainable variables from the last checkpoint.
        new_variables[name] = value

  return _create_checkpoint_from_variables(
      new_variables,
      output_dir,
      session_config=session_config)



"""Backward compatibility file, see :mod:`opennmt.optimizers.adafactor` instead."""

# pylint: disable=unused-import
from opennmt.optimizers.adafactor import AdafactorOptimizer
from opennmt.optimizers.adafactor import adafactor_decay_rate_adam
from opennmt.optimizers.adafactor import adafactor_decay_rate_pow
from opennmt.optimizers.adafactor import get_optimizer_from_params



"""Defines functions related to configuration files."""

from importlib import import_module

import os
import pickle
import sys
import tensorflow as tf
import yaml

from opennmt.models import catalog
from opennmt.utils import compat
from opennmt.utils.misc import merge_dict


def load_model_module(path):
  """Loads a model configuration file.

  Args:
    path: The relative path to the configuration file.

  Returns:
    A Python module.
  """
  dirname, filename = os.path.split(path)
  module_name, _ = os.path.splitext(filename)
  sys.path.insert(0, os.path.abspath(dirname))
  module = import_module(module_name)
  sys.path.pop(0)

  if not hasattr(module, "model"):
    raise ImportError("No model defined in {}".format(path))

  return module

def load_model_from_file(path):
  """Loads a model from a configuration file.

  Args:
    path: The relative path to the configuration file.

  Returns:
    A :class:`opennmt.models.model.Model` instance.
  """
  module = load_model_module(path)
  model = module.model()
  del sys.path_importer_cache[os.path.dirname(module.__file__)]
  del sys.modules[module.__name__]
  return model

def load_model_from_catalog(name):
  """Loads a model from the catalog.

  Args:
    name: The model name.

  Returns:
    A :class:`opennmt.models.model.Model` instance.
  """
  return getattr(catalog, name)()

def load_model(model_dir,
               model_file=None,
               model_name=None,
               serialize_model=True):
  """Loads the model from the catalog or a file.

  The model object is pickled in :obj:`model_dir` to make the model
  configuration optional for future runs.

  Args:
    model_dir: The model directory.
    model_file: An optional model configuration.
      Mutually exclusive with :obj:`model_name`.
    model_name: An optional model name from the catalog.
      Mutually exclusive with :obj:`model_file`.
    serialize_model: Serialize the model definition in the model directory.

  Returns:
    A :class:`opennmt.models.model.Model` instance.

  Raises:
    ValueError: if both :obj:`model_file` and :obj:`model_name` are set.
  """
  if model_file and model_name:
    raise ValueError("only one of model_file and model_name should be set")
  model_name_or_path = model_file or model_name
  model_description_path = os.path.join(model_dir, "model_description.py")

  # Also try to load the pickled model for backward compatibility.
  serial_model_file = os.path.join(model_dir, "model_description.pkl")

  if model_name_or_path:
    if tf.train.latest_checkpoint(model_dir) is not None:
      compat.logging.warn(
          "You provided a model configuration but a checkpoint already exists. "
          "The model configuration must define the same model as the one used for "
          "the initial training. However, you can change non structural values like "
          "dropout.")

    if model_file:
      model = load_model_from_file(model_file)
      if serialize_model:
        compat.gfile_copy(model_file, model_description_path, overwrite=True)
    elif model_name:
      model = load_model_from_catalog(model_name)
      if serialize_model:
        with compat.gfile_open(model_description_path, mode="w") as model_description_file:
          model_description_file.write("from opennmt.models import catalog\n")
          model_description_file.write("model = catalog.%s\n" % model_name)
  elif compat.gfile_exists(model_description_path):
    compat.logging.info("Loading model description from %s", model_description_path)
    model = load_model_from_file(model_description_path)
  elif compat.gfile_exists(serial_model_file):
    compat.logging.info("Loading serialized model description from %s", serial_model_file)
    with compat.gfile_open(serial_model_file, mode="rb") as serial_model:
      model = pickle.load(serial_model)
  else:
    raise RuntimeError("A model configuration is required: you probably need to "
                       "set --model or --model_type on the command line.")

  return model

def load_config(config_paths, config=None):
  """Loads configuration files.

  Args:
    config_paths: A list of configuration files.
    config: A (possibly non empty) config dictionary to fill.

  Returns:
    The configuration dictionary.
  """
  if config is None:
    config = {}

  for config_path in config_paths:
    with compat.gfile_open(config_path, mode="rb") as config_file:
      subconfig = yaml.load(config_file.read())
      # Add or update section in main configuration.
      merge_dict(config, subconfig)

  return config



"""Functions for compatibility with different TensorFlow versions."""

import tensorflow as tf


def is_tf2():
  """Returns ``True`` if running TensorFlow 2.0."""
  return tf.__version__.startswith("2")

def tf_supports(symbol):
  """Returns ``True`` if TensorFlow defines :obj:`symbol`."""
  return _string_to_tf_symbol(symbol) is not None

def tf_any(*symbols):
  """Returns the first supported symbol."""
  for symbol in symbols:
    module = _string_to_tf_symbol(symbol)
    if module is not None:
      return module
  return None

def tf_compat(v2=None, v1=None):  # pylint: disable=invalid-name
  """Returns the compatible symbol based on the current TensorFlow version.

  Args:
    v2: The candidate v2 symbol name.
    v1: The candidate v1 symbol name.

  Returns:
    A TensorFlow symbol.

  Raises:
    ValueError: if no symbol can be found.
  """
  candidates = []
  if v2 is not None:
    candidates.append(v2)
  if v1 is not None:
    candidates.append(v1)
    candidates.append("compat.v1.%s" % v1)
  symbol = tf_any(*candidates)
  if symbol is None:
    raise ValueError("Failure to resolve the TensorFlow symbol")
  return symbol

def name_from_variable_scope(name=""):
  """Creates a name prefixed by the current variable scope."""
  var_scope = tf_compat(v1="get_variable_scope")().name
  compat_name = ""
  if name:
    compat_name = "%s/" % name
  if var_scope:
    compat_name = "%s/%s" % (var_scope, compat_name)
  return compat_name

def reuse():
  """Returns ``True`` if the current variable scope is marked for reuse."""
  return tf_compat(v1="get_variable_scope")().reuse

def _string_to_tf_symbol(symbol):
  modules = symbol.split(".")
  namespace = tf
  for module in modules:
    namespace = getattr(namespace, module, None)
    if namespace is None:
      return None
  return namespace


# pylint: disable=invalid-name
gfile_copy = tf_compat(v2="io.gfile.copy", v1="gfile.Copy")
gfile_exists = tf_compat(v2="io.gfile.exists", v1="gfile.Exists")
gfile_open = tf_compat(v2="io.gfile.GFile", v1="gfile.GFile")
is_tensor = tf_compat(v2="is_tensor", v1="contrib.framework.is_tensor")
logging = tf_compat(v1="logging")
nest = tf_compat(v2="nest", v1="contrib.framework.nest")



"""RNN cells helpers."""

import collections

import tensorflow as tf


def build_cell(num_layers,
               num_units,
               mode,
               dropout=0.0,
               residual_connections=False,
               cell_class=None,
               attention_layers=None,
               attention_mechanisms=None):
  """Convenience function to build a multi-layer RNN cell.

  Args:
    num_layers: The number of layers.
    num_units: The number of units in each layer.
    mode: A ``tf.estimator.ModeKeys`` mode.
    dropout: The probability to drop units in each layer output.
    residual_connections: If ``True``, each layer input will be added to its output.
    cell_class: The inner cell class or a callable taking :obj:`num_units` as
      argument and returning a cell. Defaults to a LSTM cell.
    attention_layers: A list of integers, the layers after which to add attention.
    attention_mechanisms: A list of ``tf.contrib.seq2seq.AttentionMechanism``
      with the same length as :obj:`attention_layers`.

  Returns:
    A ``tf.nn.rnn_cell.RNNCell``.

  Raises:
    ValueError: if :obj:`attention_layers` and :obj:`attention_mechanisms` do
      not have the same length.
  """
  if cell_class is None:
    cell_class = tf.nn.rnn_cell.LSTMCell
  cells = []

  attention_mechanisms = attention_mechanisms or []
  attention_layers = attention_layers or []

  if len(attention_mechanisms) != len(attention_layers):
    raise ValueError("There must be the same number of attention mechanisms "
                     "as the number of attention layers")

  for l in range(num_layers):
    cell = cell_class(num_units)
    if l in attention_layers:
      cell = tf.contrib.seq2seq.AttentionWrapper(
          cell,
          attention_mechanisms[attention_layers.index(l)],
          attention_layer_size=num_units)
    if mode == tf.estimator.ModeKeys.TRAIN and dropout > 0.0:
      cell = tf.nn.rnn_cell.DropoutWrapper(cell, output_keep_prob=1.0 - dropout)
    if residual_connections and l > 0:
      cell = tf.nn.rnn_cell.ResidualWrapper(cell)
    cells.append(cell)

  if len(cells) == 1:
    return cells[0]
  else:
    return tf.nn.rnn_cell.MultiRNNCell(cells)

def last_encoding_from_state(state):
  """Returns the last encoding vector from the state.

  For example, this is the last hidden states of the last LSTM layer for a
  LSTM-based encoder.

  Args:
    state: The encoder state.

  Returns:
    The last encoding vector.
  """
  if isinstance(state, collections.Sequence):
    state = state[-1]
  if isinstance(state, tf.nn.rnn_cell.LSTMStateTuple):
    return state.h
  return state
